# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Basic abstract and concrete sets."""

from __future__ import absolute_import, division, print_function

from builtins import int, object
from numbers import Complex, Integral, Real

import numpy as np
from past.types.basestring import basestring

from odl.core.util import is_int_dtype, is_numeric_dtype, is_real_dtype, unique

__all__ = ('Set', 'EmptySet', 'UniversalSet', 'Field', 'Integers',
           'RealNumbers', 'ComplexNumbers', 'Strings', 'CartesianProduct',
           'SetUnion', 'SetIntersection', 'FiniteSet')


class Set(object):

    """An abstract set.

    **Abstract Methods**

    Each subclass of `Set` must implement two methods: one to
    check if an object is contained in the set and one to test if two
    sets are equal.

    **Membership test:** ``__contains__(self, other)``

    Test if ``other`` is a member of this set. This function provides
    the operator overload for ``in``.

    **Parameters:**
        other :
            Object to be tested for membership

    **Returns:**
        contains : bool
            ``True`` if ``other`` is a member of this set, ``False``
            otherwise.


    **Equality test:** ``__eq__(self, other)``

    Test if ``other`` is the same set as this set, i.e. both sets are
    of the same type and contain the same elements. This function
    provides the operator overload for ``==``.

    **Parameters:**
        other :
            Object to be tested for equality.

    **Returns:**
        equals : bool
            ``True`` if both sets are of the same type and contain the
            same elements, ``False`` otherwise.

    A default implementation of the operator overload for ``!=`` via
    ``__ne__(self, other)`` is provided as ``not self.__eq__(other)``.

    **Element creation (optional)**: ``element(self, inp=None)``

    Create an element of this set, either from scratch or from an
    input parameter.

    **Parameters:**
        inp : optional
            Object from which to create the new element

    **Returns:**
        element : member of this set
            If ``inp`` is None, return an arbitrary element.
            Otherwise, return the element created from ``inp``.
    """

    def __contains__(self, other):
        """Return ``other in self``."""
        raise NotImplementedError('abstract method')

    def contains_set(self, other):
        """Test if ``other`` is a subset of this set.

        This is a default implementation that simply tests for equality.
        It should be overridden by subclasses.

        Returns
        -------
        set_contained : bool
            ``True`` if ``other`` is contained in this set, ``False``
            otherwise.
        """
        return self == other

    def contains_all(self, other):
        """Test if all elements in ``other`` are contained in this set.

        This is a default implementation that assumes ``other`` to be
        a sequence and tests each elment of ``other`` sequentially.
        This method should be overridden by subclasses.

        Returns
        -------
        all_contained : bool
            ``True`` if all elements of ``other`` are contained in this
            set, ``False`` otherwise
        """
        return all(x in self for x in other)

    def __eq__(self, other):
        """Return ``self == other``."""
        raise NotImplementedError('abstract method')

    def __ne__(self, other):
        """Return ``self != other``."""
        return not self.__eq__(other)

    def element(self, inp=None):
        """Return an element from ``inp`` or from scratch.

        This method should be overridden by subclasses.
        """
        raise NotImplementedError('`element` method not implemented')

    @property
    def examples(self):
        """Generator creating name-value pairs of set elements.

        This method is mainly intended for diagnostics and yields elements,
        either a finite number of times or indefinitely.

        This default implementation returns
        ``('element()', self.element())`` and should be overridden by
        subclasses.
        """
        yield ('element()', self.element())

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}()'.format(self.__class__.__name__)

    def __str__(self):
        """Return ``str(self)``."""
        return '{}'.format(self.__class__.__name__)


class EmptySet(Set):

    """Set with no member elements (except ``None``).

    ``None`` is considered as "no element", i.e. ``None in EmptySet()``
    is the only test that evaluates to ``True``.
    """

    def __contains__(self, other):
        """Return ``other in self``, always ``False`` except for ``None``."""
        return other is None

    def contains_set(self, other):
        """Return ``True`` for the empty set, ``False`` otherwise."""
        return isinstance(other, EmptySet)

    def __eq__(self, other):
        """Return ``self == other``."""
        return isinstance(other, EmptySet)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash(type(self))

    def element(self, inp=None):
        """Return None."""
        return None


class UniversalSet(Set):

    """Set of all objects.

    Forget about set theory for a moment :-).
    """

    def __contains__(self, other):
        """Return ``other in self``, always ``True``."""
        return True

    def contains_set(self, other):
        """Return ``True`` for any set."""
        return isinstance(other, Set)

    def __eq__(self, other):
        """Return ``self == other``."""
        return isinstance(other, UniversalSet)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash(type(self))

    def element(self, inp=None):
        """Return ``inp`` in any case."""
        return inp


class Strings(Set):

    """Set of fixed-length (unicode) strings."""

    def __init__(self, length):
        """Initialize a new instance.

        Parameters
        ----------
        length : positive int
            Fixed length of the strings in this set.
        """
        length, length_in = int(length), length
        if length <= 0:
            raise ValueError('`length` must be positive, got {}'
                             ''.format(length_in))
        self.__length = length

    @property
    def length(self):
        """Fixed length of the strings in this set."""
        return self.__length

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        contained : bool
            ``True`` if ``other`` is a string of exactly `length`
            characters, ``False`` otherwise.
        """
        return isinstance(other, basestring) and len(other) == self.length

    def contains_all(self, other):
        """Return ``True`` if all strings in ``other`` have size `length`."""
        dtype = getattr(other, 'dtype', None)
        if dtype is None:
            dtype = np.result_type(*other)
        dtype_str = np.dtype('S{}'.format(self.length))
        dtype_uni = np.dtype('<U{}'.format(self.length))
        return dtype in (dtype_str, dtype_uni)

    def __eq__(self, other):
        """Return ``self == other``."""
        return isinstance(other, Strings) and other.length == self.length

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((type(self), self.length))

    def element(self, inp=None):
        """Return an element from ``inp`` or from scratch."""
        if inp is not None:
            s = str(inp)[:self.length]
            s += ' ' * (self.length - len(s))
            return s
        else:
            return ' ' * self.length

    @property
    def examples(self):
        """Return example strings 'hello', 'world' (size adapted)."""
        hello_str = 'hello'[:self.length]
        hello_str += ' ' * (self.length - len(hello_str))
        world_str = 'world'[:self.length]
        world_str += ' ' * (self.length - len(world_str))
        return [('hello', hello_str), ('world', world_str)]

    def __repr__(self):
        """Return ``repr(self)``."""
        return 'Strings({})'.format(self.length)


class Field(Set):

    """A set that satisfies the field axioms.

    Examples: `RealNumbers`, `ComplexNumbers` or
    the finite field :math:`F_2`.

    See `the Wikipedia entry on fields
    <https://en.wikipedia.org/wiki/Field_%28mathematics%29>`_ for
    further information.
    """

    @property
    def field(self):
        """Field of scalars for a field is itself.

        Notes
        -----
        This is a hack to make fields to work via duck-typing with
        `LinearSpace`'s.
        """
        return self


class ComplexNumbers(Field):

    """Set of complex numbers."""

    def __contains__(self, other):
        """Return ``other in self``."""
        return isinstance(other, Complex)

    def contains_set(self, other):
        """Return ``True`` if ``other`` is a subset of the complex numbers.

        Returns
        -------
        contained : bool
            ``True`` if  ``other`` is an instance of `ComplexNumbers`,
            `RealNumbers` or `Integers`, ``False`` otherwise.

        Examples
        --------
        >>> complex_numbers = ComplexNumbers()
        >>> complex_numbers.contains_set(RealNumbers())
        True
        """
        if other is self:
            return True

        return (isinstance(other, ComplexNumbers) or
                isinstance(other, RealNumbers) or
                isinstance(other, Integers))

    def contains_all(self, other):
        """Return ``True`` if ``other`` is a sequence of complex numbers."""
        dtype = getattr(other, 'dtype', None)
        if dtype is None:
            dtype = np.result_type(*other)
        return is_numeric_dtype(dtype)

    def __eq__(self, other):
        """Return ``self == other``."""
        if other is self:
            return True

        return isinstance(other, ComplexNumbers)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash(type(self))

    def element(self, inp=None):
        """Return a complex number from ``inp`` or from scratch."""
        if inp is None:
            return complex(0.0, 0.0)
        else:
            return complex(inp)

    @property
    def examples(self):
        """Return examples of complex numbers."""
        numbers = [-1.0, 0.5, 0.0 + 2.0j, 0.0, 0.01, 1.0 + 1.0j, 1.0j, 1.0]
        return [(str(x), x) for x in numbers]


class RealNumbers(Field):

    """Set of real numbers."""

    def __contains__(self, other):
        """Return ``other in self``."""
        return isinstance(other, Real)

    def contains_set(self, other):
        """Return ``True`` if ``other`` is a subset of the real numbers.

        Returns
        -------
        contained : bool
            ``True`` if other is an instance of `RealNumbers` or
            `Integers` False otherwise.

        Examples
        --------
        >>> real_numbers = RealNumbers()
        >>> real_numbers.contains_set(RealNumbers())
        True
        """
        if other is self:
            return True

        return (isinstance(other, RealNumbers) or
                isinstance(other, Integers))

    def contains_all(self, array):
        """Test if `array` is an array of real numbers."""
        dtype = getattr(array, 'dtype', None)
        if dtype is None:
            dtype = np.result_type(*array)
        return is_real_dtype(dtype)

    def __eq__(self, other):
        """Return ``self == other``."""
        if other is self:
            return True

        return isinstance(other, RealNumbers)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash(type(self))

    def element(self, inp=None):
        """Return a real number from ``inp`` or from scratch."""
        if inp is not None:
            return float(inp)
        else:
            return 0.0

    @property
    def examples(self):
        """Return examples of real numbers."""
        numbers = [-1.0, 0.5, 0.0, 0.01, 1.0]
        return [(str(x), x) for x in numbers]


class Integers(Set):

    """Set of integers."""

    def __contains__(self, other):
        """Return ``other in self``."""
        return isinstance(other, Integral)

    def contains_set(self, other):
        """Test if ``other`` is a subset of the integers.

        Returns
        -------
        contained : bool
            ``True`` if other is an instance of `Integers`,
            ``False`` otherwise.

        Examples
        --------
        >>> integers = Integers()
        >>> integers.contains_set(RealNumbers())
        False
        """
        if other is self:
            return True

        return isinstance(other, Integers)

    def contains_all(self, other):
        """Return ``True`` if ``other`` is a sequence of integers."""
        dtype = getattr(other, 'dtype', None)
        if dtype is None:
            dtype = np.result_type(*other)
        return is_int_dtype(dtype)

    def __eq__(self, other):
        """Return ``self == other``."""
        if other is self:
            return True

        return isinstance(other, Integers)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash(type(self))

    def element(self, inp=None):
        """Return an integer from ``inp`` or from scratch."""
        if inp is not None:
            return int(inp)
        else:
            return 0

    @property
    def examples(self):
        """Return examples of integers."""
        numbers = [-1, 0, 1]
        return [(str(x), x) for x in numbers]


class CartesianProduct(Set):

    """Cartesian product of a finite number of sets.

    The elements of this set are tuples where the i-th entry
    is an element of the i-th set.
    """

    def __init__(self, *sets):
        """Initialize a new instance."""
        for set_ in sets:
            if not isinstance(set_, Set):
                raise TypeError('{!r} is not a Set instance.'.format(set_))

        self.__sets = tuple(sets)

    @property
    def sets(self):
        """The sets of this cartesian product as a tuple."""
        return self.__sets

    def __contains__(self, other):
        """Return ``other in self``.

        Parameters
        ----------
        other :
            Object to be tested for membership

        Returns
        -------
        contains : bool
            ``True`` if ``other`` is a sequence with same length as this
            Cartesian product, and each entry is contained in the set with
            corresponding index, ``False`` otherwise.
        """
        try:
            len(other)
        except TypeError:
            return False
        return (len(other) == len(self) and
                all(p in set_ for set_, p in zip(self.sets, other)))

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `CartesianProduct` instance,
            has the same length as this Cartesian product and all sets
            with the same index are equal, ``False`` otherwise.
        """
        return (type(self) == type(other) and
                self.sets == other.sets)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((type(self), self.sets))

    def element(self, inp=None):
        """Create a `CartesianProduct` element.

        Parameters
        ----------
        inp : iterable, optional
            Collection of input values for the
            `LinearSpace.element` methods
            of all sets in the Cartesian product.

        Returns
        -------
        element : tuple
            A tuple of the given input
        """
        if inp is None:
            tpl = tuple(set_.element() for set_ in self.sets)
        else:
            tpl = tuple(set_.element(inpt)
                        for inpt, set_ in zip(inp, self.sets))

            if len(tpl) != len(self):
                raise ValueError('input provides only {} values, needed '
                                 'are {}'.format(len(tpl), len(self)))

        return tpl

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.sets)

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> prod = odl.CartesianProduct(reals, complexnrs, complexnrs, reals)
        >>> prod[1]
        ComplexNumbers()
        >>> prod[2:4]
        CartesianProduct(ComplexNumbers(), RealNumbers())
        """
        if isinstance(indices, slice):
            return CartesianProduct(*self.sets[indices])
        else:
            return self.sets[indices]

    def __str__(self):
        """Return ``str(self)``."""
        return ' x '.join(str(set_) for set_ in self.sets)

    def __repr__(self):
        """Return ``repr(self)``."""
        sets_str = ', '.join(repr(set_) for set_ in self.sets)
        return '{}({})'.format(self.__class__.__name__, sets_str)


class SetUnion(Set):

    """The union of several subsets.

    The elements of this set are elements of at least one of the subsets.

    This is a *lazy* union, i.e. there is no intelligence and the set is
    literally stored as the union of its subsets.
    """

    def __init__(self, *sets):
        """Initialize a new instance.

        Parameters
        ----------
        set1, ..., setN : `Set`
            The sets whose union should be taken.
            Any duplicates are ignored.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> union = odl.SetUnion(reals, complexnrs)
        """
        for set_ in sets:
            if not isinstance(set_, Set):
                raise TypeError('{!r} is not a Set instance.'.format(set_))

        self.__sets = tuple(unique(sets))

    @property
    def sets(self):
        """The sets of this union as a tuple."""
        return self.__sets

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        contains : bool
            ``True`` if ``other`` is a member of any subset,
            ``False`` otherwise.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> union = odl.SetUnion(reals, complexnrs)
        >>> 2 + 1j in union
        True
        >>> [1, 2] in union
        False
        """
        return any(other in set for set in self.sets)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `SetUnion` instance, and
            has the same subsets as this set, ``False`` otherwise.
        """
        return (type(self) == type(other) and
                all(set_ in other for set_ in self) and
                all(set_ in self for set_ in other))

    def __hash__(self):
        """Return ``hash(self)``."""
        # Use `set` to allow permutations
        return hash((type(self), set(self.sets)))

    def element(self, inp=None):
        """Create a new element.

        First tries calling the first set, then the second, etc.

        For more specific control, use ``set[i].element()`` to pick which
        subset to use.
        """
        for set in self.sets:
            try:
                return set.element(inp)
            except NotImplementedError:
                pass
        raise NotImplementedError('`element` not implemented for any of the '
                                  'subsets')

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.sets)

    def __getitem__(self, indcs):
        """Return ``self[indcs]``.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> union = odl.SetUnion(reals, complexnrs)
        >>> union[0]
        RealNumbers()
        >>> union[:]
        SetUnion(RealNumbers(), ComplexNumbers())
        """
        if isinstance(indcs, slice):
            return SetUnion(*self.sets[indcs])
        else:
            return self.sets[indcs]

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> odl.SetUnion(reals, complexnrs)
        SetUnion(RealNumbers(), ComplexNumbers())
        """
        sets_str = ', '.join(repr(set_) for set_ in self.sets)
        return '{}({})'.format(self.__class__.__name__, sets_str)


class SetIntersection(Set):

    """The intersection of several subsets.

    The elements of this set are elements of all the subsets.

    This is a *lazy* intersection, i.e. there is no intelligence and the set is
    literally stored as the intersection of its subsets.
    """

    def __init__(self, *sets):
        """Initialize a new instance.

        Parameters
        ----------
        set1, ..., setN : `Set`
            The sets whose intersection should be taken.
            Any duplicates are ignored.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> union = odl.SetIntersection(reals, complexnrs)
        """
        for set_ in sets:
            if not isinstance(set_, Set):
                raise TypeError('{!r} is not a Set instance.'.format(set_))

        self.__sets = tuple(unique(sets))

    @property
    def sets(self):
        """The sets of this intersection as a tuple."""
        return self.__sets

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        contains : bool
            ``True`` if ``other`` is a member of all subsets,
            ``False`` otherwise.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> intersection = odl.SetIntersection(reals, complexnrs)
        >>> 1.0 in intersection
        True
        >>> 1.0j in intersection
        False
        """
        return all(other in set for set in self.sets)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `SetUnion` instance, and
            has the same subsets as this set, ``False`` otherwise.
        """
        return (type(self) == type(other) and
                all(set_ in other for set_ in self) and
                all(set_ in self for set_ in other))

    def __hash__(self):
        """Return ``hash(self)``."""
        # Use `set` to allow permutations
        return hash((type(self), set(self.sets)))

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.sets)

    def __getitem__(self, indcs):
        """Return ``self[indcs]``.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> intersection = odl.SetIntersection(reals, complexnrs)
        >>> intersection[0]
        RealNumbers()
        >>> intersection[:]
        SetIntersection(RealNumbers(), ComplexNumbers())
        """
        if isinstance(indcs, slice):
            return SetIntersection(*self.sets[indcs])
        else:
            return self.sets[indcs]

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> reals, complexnrs = odl.RealNumbers(), odl.ComplexNumbers()
        >>> odl.SetIntersection(reals, complexnrs)
        SetIntersection(RealNumbers(), ComplexNumbers())
        """
        sets_str = ', '.join(repr(set_) for set_ in self.sets)
        return '{}({})'.format(self.__class__.__name__, sets_str)


class FiniteSet(Set):

    """A set given by a finite number of elements."""

    def __init__(self, *elements):
        """Initialize a new instance.

        Parameters
        ----------
        element1, ..., elementN : `Set`
            The elements in the set. Any duplicates are ignored.

        Examples
        --------
        >>> set = odl.FiniteSet(1, 'string')
        """
        self.__elements = tuple(unique(elements))

    @property
    def elements(self):
        """The elements as a tuple."""
        return self.__elements

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        contains : bool
            ``True`` if ``other`` is an element in `elements`,
            ``False`` otherwise.

        Examples
        --------
        >>> set = odl.FiniteSet(1, 'string')
        >>> 1 in set
        True
        >>> 2 in set
        False
        >>> 'string' in set
        True
        """
        return other in self.elements

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `SetUnion` instance, and
            has the same subsets as this set, ``False`` otherwise.
        """
        # Need to loop since order could be different
        return (type(self) == type(other) and
                all(el in other for el in self) and
                all(el in self for el in other))

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((type(self), set(self.elements)))

    def element(self, inp=None):
        """Create a new element.

        For more specific control, use set[i].element() to pick which subset to
        use.
        """
        if inp is None:
            return self.elements[0]
        elif inp in self.elements:
            return inp
        else:
            raise ValueError('cannot convert inp {} to element in {}'
                             ''.format(inp, self))

    def __getitem__(self, indcs):
        """Return ``self[indcs]``.

        Examples
        --------
        >>> set = odl.FiniteSet(1, 2, 3, 'string')
        >>> set[:3]
        FiniteSet(1, 2, 3)
        >>> set[3]
        'string'
        """
        if isinstance(indcs, slice):
            return FiniteSet(*self.elements[indcs])
        else:
            return self.elements[indcs]

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> odl.FiniteSet(1, 'string')
        FiniteSet(1, 'string')
        """
        elements_str = ', '.join(repr(el) for el in self.elements)
        return '{}({})'.format(self.__class__.__name__, elements_str)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Method(s) to find optimal reconstruction parameter(s) w.r.t. given FOM."""

import numpy as np
import scipy.optimize

__all__ = ('optimal_parameters', )


def optimal_parameters(reconstruction, fom, phantoms, data,
                       initial=None, univariate=False):
    r"""Find the optimal parameters for a reconstruction method.

    Notes
    -----
    For a forward operator :math:`A : X \to Y`, a reconstruction operator
    parametrized by :math:`\theta` is some operator
    :math:`R_\theta : Y \to X`
    such that

    .. math::
        R_\theta(A(x)) \approx x.

    The optimal choice of :math:`\theta` is given by

    .. math::
        \theta = \arg\min_\theta fom(R(A(x) + noise), x)

    where :math:`fom : X \times X \to \mathbb{R}` is a figure of merit.

    Parameters
    ----------
    reconstruction : callable
        Function that takes two parameters:

            * data : The data to be reconstructed
            * parameters : Parameters of the reconstruction method

        The function should return the reconstructed image.
    fom : callable
        Function that takes two parameters:

            * reconstructed_image
            * true_image

        and returns a scalar figure of merit.
    phantoms : sequence of ODL Tensors
        True images.
    data : sequence
        The data to reconstruct from.
    initial : array-like or pair
        Initial guess for the parameters. It is
        - a required array in the multivariate case
        - an optional pair in the univariate case.
    univariate : bool, optional
        Whether to use a univariate solver

    Returns
    -------
    parameters : 'numpy.ndarray'
        The  optimal parameters for the reconstruction problem.
    """

    def func(lam):
        # Function to be minimized by scipy
        return sum(fom(reconstruction(datai, lam), phantomi)
                   for phantomi, datai in zip(phantoms, data))

    # Pick resolution to fit the one used by the space
    tol = np.finfo(phantoms[0].space.dtype_identifier).resolution * 10

    if univariate:
        # We use a faster optimizer for the one parameter case
        result = scipy.optimize.minimize_scalar(
            func, bracket=initial, tol=tol, bounds=None,
            options={'disp': False})
        return result.x
    else:
        # Use a gradient free method to find the best parameters
        initial = np.asarray(initial)
        parameters = scipy.optimize.fmin_powell(
            func, initial, xtol=tol, ftol=tol, disp=False)
        return parameters



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Setup script for ODL.

Installation command::

    pip install [--user] [-e] .
"""

from __future__ import print_function, absolute_import

from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(cmdclass={'test': PyTest})



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities for creating phantoms."""

from __future__ import print_function, division, absolute_import
import numpy as np


__all__ = ('cylinders_from_ellipses',)


def cylinders_from_ellipses(ellipses):
    """Create 3d cylinders from ellipses."""
    ellipses = np.asarray(ellipses)
    ellipsoids = np.zeros((ellipses.shape[0], 10))
    ellipsoids[:, [0, 1, 2, 4, 5, 7]] = ellipses
    ellipsoids[:, 3] = 100000.0

    return ellipsoids


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tomography helpers for Elekta systems."""

import numpy as np
import odl

__all__ = ('elekta_icon_geometry',
           'elekta_icon_space',
           'elekta_icon_fbp',
           'elekta_xvi_geometry',
           'elekta_xvi_space',
           'elekta_xvi_fbp')


def elekta_icon_geometry(sad=780.0, sdd=1000.0,
                         piercing_point=(390.0, 0.0),
                         angles=None, num_angles=None,
                         detector_shape=(780, 720)):
    """Tomographic geometry of the Elekta Icon CBCT system.

    See the [whitepaper]_ for specific descriptions of each parameter.

    All measurments are given in millimeters unless otherwise stated.

    Parameters
    ----------
    sad : float, optional
        Source to Axis distance.
    sdd : float, optional
        Source to Detector distance.
    piercing_point : sequence of foat, optional
        Position in the detector (in pixel coordinates) that a beam from the
        source, passing through the axis of rotation perpendicularly, hits.
    angles : array-like, optional
        List of angles given in radians that the projection images were taken
        at. Exclusive with num_angles.
        Default: np.linspace(1.2, 5.0, 332)
    num_angles : int, optional
        Number of angles. Exclusive with angles.
        Default: 332
    detector_shape : sequence of int, optional
        Shape of the detector (in pixels). Useful if a sub-sampled system
        should be studied.

    Returns
    -------
    elekta_icon_geometry : `ConeBeamGeometry`

    Examples
    --------
    Create default geometry:

    >>> from odl.contrib import tomo
    >>> geometry = tomo.elekta_icon_geometry()

    Use a smaller detector (improves efficiency):

    >>> small_geometry = tomo.elekta_icon_geometry(detector_shape=[100, 100])

    See Also
    --------
    elekta_icon_space : Default reconstruction space for the Elekta Icon CBCT.
    elekta_icon_fbp: Default reconstruction method for the Elekta Icon CBCT.

    References
    ----------
    .. [whitepaper] *Design and performance characteristics of a Cone Beam
       CT system for Leksell Gamma Knife Icon*
    """
    sad = float(sad)
    assert sad > 0
    sdd = float(sdd)
    assert sdd > sad
    piercing_point = np.array(piercing_point, dtype=float)
    assert piercing_point.shape == (2,)

    if angles is not None and num_angles is not None:
        raise ValueError('cannot provide both `angles` and `num_angles`')
    elif angles is not None:
        angles = odl.nonuniform_partition(angles)
        assert angles.ndim == 1
    elif num_angles is not None:
        angles = odl.uniform_partition(1.2, 5.0, num_angles)
    else:
        angles = odl.uniform_partition(1.2, 5.0, 332)

    detector_shape = np.array(detector_shape, dtype=int)

    # Constant system parameters
    pixel_size = 0.368
    det_extent_mm = np.array([287.04, 264.96])

    # Compute the detector partition
    piercing_point_mm = pixel_size * piercing_point
    det_min_pt = -piercing_point_mm
    det_max_pt = det_min_pt + det_extent_mm
    detector_partition = odl.uniform_partition(min_pt=det_min_pt,
                                               max_pt=det_max_pt,
                                               shape=detector_shape)

    # Create the geometry
    geometry = odl.applications.tomo.ConeBeamGeometry(
        angles, detector_partition,
        src_radius=sad, det_radius=sdd - sad)

    return geometry


def elekta_icon_space(shape=(448, 448, 448), **kwargs):
    """Default reconstruction space for the Elekta Icon CBCT.

    See the [whitepaper]_ for further information.

    Parameters
    ----------
    shape : sequence of int, optional
        Shape of the space, in voxels.
    kwargs :
        Keyword arguments to pass to `uniform_discr` to modify the space, e.g.
        use another backend. By default, the dtype is set to float32.

    Returns
    -------
    elekta_icon_space : `DiscretizedSpace`

    Examples
    --------
    Create default space:

    >>> from odl.contrib import tomo
    >>> space = tomo.elekta_icon_space()

    Create sub-sampled space:

    >>> space = tomo.elekta_icon_space(shape=(100, 100, 100))

    See Also
    --------
    elekta_icon_geometry: Geometry for the Elekta Icon CBCT.
    elekta_icon_fbp: Default reconstruction method for the Elekta Icon CBCT.

    References
    ----------
    .. [whitepaper] *Design and performance characteristics of a Cone Beam
       CT system for Leksell Gamma Knife Icon*
    """
    if 'dtype' not in kwargs:
        kwargs['dtype'] = 'float32'
    return odl.uniform_discr(min_pt=[-112.0, -112.0, 0.0],
                             max_pt=[112.0, 112.0, 224.0],
                             shape=shape,
                             **kwargs)


def elekta_icon_fbp(ray_transform,
                    padding=False, filter_type='Hann', frequency_scaling=0.6,
                    parker_weighting=True):
    """Approximation of the FDK reconstruction used in the Elekta Icon.

    Parameters
    ----------
    ray_transform : `RayTransform`
        The ray transform to be used, should have an Elekta Icon geometry.
    padding : bool, optional
        Whether the FBP filter should use padding, increases memory use
        significantly.
    filter_type : str, optional
        Type of filter to apply in the FBP filter.
    frequency_scaling : float, optional
        Frequency scaling for FBP filter.
    parker_weighting : bool, optional
        Whether Parker weighting should be applied to compensate for partial
        scan.

    Returns
    -------
    elekta_icon_fbp : `DiscretizedSpace`

    Examples
    --------
    Create default FBP for default geometry:

    >>> from odl.contrib import tomo
    >>> geometry = tomo.elekta_icon_geometry()
    >>> space = tomo.elekta_icon_space()
    >>> ray_transform = odl.applications.tomo.RayTransform(space, geometry)
    >>> fbp_op = tomo.elekta_icon_fbp(ray_transform)
    """
    fbp_op = odl.applications.tomo.fbp_op(ray_transform,
                             padding=padding,
                             filter_type=filter_type,
                             frequency_scaling=frequency_scaling)
    if parker_weighting:
        parker_weighting = odl.applications.tomo.parker_weighting(ray_transform)
        fbp_op = fbp_op * parker_weighting

    return fbp_op


def elekta_xvi_geometry(sad=1000.0, sdd=1500.0,
                        piercing_point=(512.0, 512.0),
                        angles=None, num_angles=None,
                        detector_shape=(1024, 1024)):
    """Tomographic geometry of the Elekta XVI system.

    All measurments are given in millimeters unless otherwise stated.

    Parameters
    ----------
    sad : float, optional
        Source to Axis distance.
    sdd : float, optional
        Source to Detector distance.
    piercing_point : sequence of foat, optional
        Position in the detector (in pixel coordinates) that a beam from the
        source, passing through the axis of rotation perpendicularly, hits.
    angles : array-like, optional
        List of angles given in radians that the projection images were taken
        at. Exclusive with num_angles.
        Default: np.linspace(0, 2 * np.pi, 650, endpoint=False)
    num_angles : int, optional
        Number of angles. Exclusive with angles.
        Default: 332
    detector_shape : sequence of int, optional
        Shape of the detector (in pixels). Useful if a sub-sampled system
        should be studied.

    Returns
    -------
    elekta_xvi_geometry : `ConeBeamGeometry`

    Examples
    --------
    Create default geometry:

    >>> from odl.contrib import tomo
    >>> geometry = tomo.elekta_xvi_geometry()

    Use a smaller detector (improves efficiency):

    >>> small_geometry = tomo.elekta_xvi_geometry(detector_shape=[100, 100])

    See Also
    --------
    elekta_xvi_space : Default reconstruction space for the Elekta XVI CBCT.
    elekta_xvi_fbp: Default reconstruction method for the Elekta XVI CBCT.
    """
    sad = float(sad)
    assert sad > 0
    sdd = float(sdd)
    assert sdd > sad
    piercing_point = np.array(piercing_point, dtype=float)
    assert piercing_point.shape == (2,)

    if angles is not None and num_angles is not None:
        raise ValueError('cannot provide both `angles` and `num_angles`')
    elif angles is not None:
        angles = odl.nonuniform_partition(angles)
        assert angles.ndim == 1
    elif num_angles is not None:
        angles = odl.uniform_partition(0, 2 * np.pi, num_angles)
    else:
        angles = odl.uniform_partition(0, 2 * np.pi, 650)

    detector_shape = np.array(detector_shape, dtype=int)

    # Constant system parameters
    pixel_size = 0.4
    det_extent_mm = np.array([409.6, 409.6])

    # Compute the detector partition
    piercing_point_mm = pixel_size * piercing_point
    det_min_pt = -piercing_point_mm
    det_max_pt = det_min_pt + det_extent_mm
    detector_partition = odl.uniform_partition(min_pt=det_min_pt,
                                               max_pt=det_max_pt,
                                               shape=detector_shape)

    # Create the geometry
    geometry = odl.applications.tomo.ConeBeamGeometry(
        angles, detector_partition,
        src_radius=sad, det_radius=sdd - sad)

    return geometry


def elekta_xvi_space(shape=(512, 512, 512), **kwargs):
    """Default reconstruction space for the Elekta XVI CBCT.

    Parameters
    ----------
    shape : sequence of int, optional
        Shape of the space, in voxels.
    kwargs :
        Keyword arguments to pass to `uniform_discr` to modify the space, e.g.
        use another backend. By default, the dtype is set to float32.

    Returns
    -------
    elekta_xvi_space : `DiscretizedSpace`

    Examples
    --------
    Create default space:

    >>> from odl.contrib import tomo
    >>> space = tomo.elekta_xvi_space()

    Create sub-sampled space:

    >>> space = tomo.elekta_xvi_space(shape=(100, 100, 100))

    See Also
    --------
    elekta_xvi_geometry: Geometry for the Elekta XVI CBCT.
    elekta_xvi_fbp: Default reconstruction method for the Elekta XVI CBCT.
    """
    if 'dtype' not in kwargs:
        kwargs['dtype'] = 'float32'
    return odl.uniform_discr(min_pt=[-128.0, -128, -128.0],
                             max_pt=[128.0, 128.0, 128.0],
                             shape=shape,
                             **kwargs)


def elekta_xvi_fbp(ray_transform,
                   padding=False, filter_type='Hann', frequency_scaling=0.6):
    """Approximation of the FDK reconstruction used in the Elekta XVI.

    Parameters
    ----------
    ray_transform : `RayTransform`
        The ray transform to be used, should have an Elekta XVI geometry.
    padding : bool, optional
        Whether the FBP filter should use padding, increases memory use
        significantly.
    filter_type : str, optional
        Type of filter to apply in the FBP filter.
    frequency_scaling : float, optional
        Frequency scaling for FBP filter.

    Returns
    -------
    elekta_xvi_fbp : `DiscretizedSpace`

    Examples
    --------
    Create default FBP for default geometry:

    >>> from odl.contrib import tomo
    >>> geometry = tomo.elekta_xvi_geometry()
    >>> space = tomo.elekta_xvi_space()
    >>> ray_transform = odl.applications.tomo.RayTransform(space, geometry)
    >>> fbp_op = tomo.elekta_xvi_fbp(ray_transform)
    """
    fbp_op = odl.applications.tomo.fbp_op(ray_transform,
                             padding=padding,
                             filter_type=filter_type,
                             frequency_scaling=frequency_scaling)

    return fbp_op


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import pytest

import odl
from odl.core.set.sets import (EmptySet, UniversalSet, Strings, ComplexNumbers,
                          RealNumbers, Integers)


def test_empty_set():
    X = EmptySet()
    Z = Integers()

    # __contains
    assert None in X
    assert 1 not in X

    # Contains_set
    assert X.contains_set(X)
    assert not X.contains_set(Z)

    # __eq__
    assert X == X
    assert X != Z

    # element
    assert X.element() is None


def test_universal_set():
    X = UniversalSet()
    Z = Integers()

    # __contains
    assert None in X
    assert 1 in X

    # Contains_set
    assert X.contains_set(X)
    assert X.contains_set(Z)
    assert not X.contains_set(1)

    # __eq__
    assert X == X
    assert X != Z

    # element
    assert X.element() is None
    assert X.element(1) == 1


def test_strings():
    S1 = Strings(1)
    S6 = Strings(6)
    Z = Integers()

    with pytest.raises(ValueError):  # Raises in the conversion to int
        Strings('fail')

    with pytest.raises(ValueError):
        Strings(-2)

    # __contains
    assert 'c' in S1
    assert 'string' not in S1
    assert 'c' not in S6
    assert 'string' in S6

    assert 1 not in S1
    assert 1 not in S6

    # __eq__
    assert S1 == S1
    assert S1 != S6
    assert S6 != S1
    assert S6 == S6
    assert S1 != Z

    # element
    assert S1.element() == ' '
    assert S6.element() == ' ' * 6
    assert S1.element('') == ' '
    assert S6.element('') == ' ' * 6
    assert S1.element('a') == 'a'
    assert S1.element('abcdefg') == 'a'
    assert S6.element('a') == 'a     '
    assert S6.element('abcdefg') == 'abcdef'


def test_complex():
    C = ComplexNumbers()
    R = RealNumbers()
    Z = Integers()

    # __contains__
    assert -1 in C
    assert 1 in C
    assert 0 in C
    assert -1.0 in C
    assert 1.0 in C
    assert 0.0 in C
    assert 2j in C
    assert 2 + 2j in C

    assert 'a' not in C

    # contains_set
    assert C.contains_set(C)
    assert C.contains_set(R)
    assert C.contains_set(Z)

    # __eq__
    assert C == C
    assert C != R
    assert C != Z

    # element
    assert C.element() == complex(0.0, 0.0)
    assert C.element(1) == complex(1.0, 0.0)
    assert C.element(1 + 2j) == complex(1.0, 2.0)


def test_real():
    C = ComplexNumbers()
    R = RealNumbers()
    Z = Integers()

    # __contains__
    assert -1 in R
    assert 1 in R
    assert 0 in R
    assert -1.0 in R
    assert 1.0 in R
    assert 0.0 in R

    assert 2j not in R
    assert 2 + 2j not in R
    assert 'a' not in R

    # contains_set
    assert not R.contains_set(C)
    assert R.contains_set(R)
    assert C.contains_set(Z)

    # __eq__
    assert R != C
    assert R == R
    assert R != Z

    # element
    assert C.element() == float(0.0)
    assert C.element(1) == float(1.0)


def test_integers():
    C = ComplexNumbers()
    R = RealNumbers()
    Z = Integers()

    # __contains__
    assert -1 in Z
    assert 1 in Z
    assert 0 in Z

    assert -1.0 not in Z
    assert 1.0 not in Z
    assert 0.0 not in Z
    assert 2j not in Z
    assert 2 + 2j not in Z
    assert 'a' not in Z

    # contains_set
    assert not Z.contains_set(C)
    assert not Z.contains_set(R)
    assert Z.contains_set(Z)

    # __eq__
    assert Z != C
    assert Z != R
    assert Z == Z

    # element
    assert Z.element() == int(0)
    assert Z.element(1) == int(1)
    assert Z.element(1.5) == int(1.5)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Contributed code for the stochastic PDHG.

[CERS2017] A. Chambolle, M. J. Ehrhardt, P. Richtarik and C.-B. Schoenlieb,
*Stochastic Primal-Dual Hybrid Gradient Algorithm with Arbitrary Sampling
and Imaging Applications*. ArXiv: http://arxiv.org/abs/1706.04957 (2017).

[E+2017] M. J. Ehrhardt, P. J. Markiewicz, P. Richtarik, J. Schott,
A. Chambolle and C.-B. Schoenlieb, *Faster PET reconstruction with a
stochastic primal-dual hybrid gradient method*. Wavelets and Sparsity XVII,
58 (2017) http://doi.org/10.1117/12.2272946.
"""

from __future__ import absolute_import

__all__ = ()

from .misc import *
__all__ += misc.__all__

from .stochastic_primal_dual_hybrid_gradient import *
__all__ += stochastic_primal_dual_hybrid_gradient.__all__



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Nonlinear version of the conjugate gradient method."""

from __future__ import print_function, division, absolute_import

from odl.solvers.util import ConstantLineSearch


__all__ = ('conjugate_gradient_nonlinear',)


def conjugate_gradient_nonlinear(f, x, line_search=1.0, maxiter=1000, nreset=0,
                                 tol=1e-16, beta_method='FR',
                                 callback=None):
    r"""Conjugate gradient for nonlinear problems.

    Parameters
    ----------
    f : `Functional`
        Functional with ``f.gradient``.
    x : ``op.domain`` element
        Vector to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
    line_search : float or `LineSearch`, optional
        Strategy to choose the step length. If a float is given, it is used as
        a fixed step length.
    maxiter : int, optional
        Maximum number of iterations to perform.
    nreset : int, optional
        Number of times the solver should be reset. Default: no reset.
    tol : float, optional
        Tolerance that should be used to terminating the iteration.
    beta_method : {'FR', 'PR', 'HS', 'DY'}, optional
        Method to calculate ``beta`` in the iterates.

        - ``'FR'`` : Fletcher-Reeves
        - ``'PR'`` : Polak-Ribiere
        - ``'HS'`` : Hestenes-Stiefel
        - ``'DY'`` : Dai-Yuan

    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate.

    Notes
    -----
    This is a general and optimized implementation of the nonlinear conjguate
    gradient method for solving a general unconstrained optimization problem

    .. math::
        \min f(x)

    for a differentiable functional
    :math:`f: \mathcal{X}\to \mathbb{R}` on a Hilbert space
    :math:`\mathcal{X}`. It does so by finding a zero of the gradient

    .. math::
        \nabla f: \mathcal{X} \to \mathcal{X}.

    The method is described in a
    `Wikipedia article
    <https://en.wikipedia.org/wiki/Nonlinear_conjugate_gradient_method>`_.

    See Also
    --------
    odl.solvers.smooth.newton.bfgs_method :
        Quasi-Newton solver for the same problem
    odl.solvers.iterative.iterative.conjugate_gradient :
        Optimized solver for least-squares problem with linear and symmetric
        operator
    odl.solvers.iterative.iterative.conjugate_gradient_normal :
        Equivalent solver but for least-squares problem with linear operator
    """
    if x not in f.domain:
        raise TypeError('`x` {!r} is not in the domain of `f` {!r}'
                        ''.format(x, f.domain))

    if not callable(line_search):
        line_search = ConstantLineSearch(line_search)

    if beta_method not in ['FR', 'PR', 'HS', 'DY']:
        raise ValueError('unknown ``beta_method``')

    for _ in range(nreset + 1):
        # First iteration is done without beta
        dx = -f.gradient(x)
        dir_derivative = -dx.inner(dx)
        if abs(dir_derivative) < tol:
            return
        a = line_search(x, dx, dir_derivative)
        x.lincomb(1, x, a, dx)  # x = x + a * dx

        s = dx  # for 'HS' and 'DY' beta methods

        for _ in range(maxiter // (nreset + 1)):
            # Compute dx as -grad f
            dx, dx_old = -f.gradient(x), dx

            # Calculate "beta"
            if beta_method == 'FR':
                beta = dx.inner(dx) / dx_old.inner(dx_old)
            elif beta_method == 'PR':
                beta = dx.inner(dx - dx_old) / dx_old.inner(dx_old)
            elif beta_method == 'HS':
                beta = - dx.inner(dx - dx_old) / s.inner(dx - dx_old)
            elif beta_method == 'DY':
                beta = - dx.inner(dx) / s.inner(dx - dx_old)
            else:
                raise RuntimeError('unknown ``beta_method``')

            # Reset beta if negative.
            beta = max(0, beta)

            # Update search direction
            s.lincomb(1, dx, beta, s)  # s = dx + beta * s

            # Find optimal step along s
            dir_derivative = -dx.inner(s)

            if abs(dir_derivative) <= tol:
                return
            a = line_search(x, s, dir_derivative)

            # Update position
            x.lincomb(1, x, a, s)  # x = x + a * s

            if callback is not None:
                callback(x)




from odl.core.sparse.sparse_template import SparseMatrixFormat, _registered_sparse_formats


from typing import Optional


IS_INITIALIZED = False

def _initialize_if_needed():
    """Initialize ``_registered_sparse_formats`` if not already done."""
    global IS_INITIALIZED
    if not IS_INITIALIZED:
        import odl.backends.sparse.scipy_backend
        import importlib.util       
        torch_module = importlib.util.find_spec("torch")
        if torch_module is not None:
            try:
                import odl.backends.sparse.pytorch_backend
            except ModuleNotFoundError:
                pass
        IS_INITIALIZED = True

class SparseMatrix():    
    """
    SparseMatrix is the ODL interface to the sparse Matrix supports in different backends.

    Note:
    The user is responsible for using the *args and **kwargs expected by the respective backends:
    Pytorch: 
        -> COO: https://docs.pytorch.org/docs/stable/generated/torch.sparse_coo_tensor.html
    Scipy: 
        -> COO: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.coo_matrix.
        
    Examples:
    SparseMatrix('COO', 'pytorch', 
                [[0, 1, 1],[2, 0, 2]], [3, 4, 5], 
                device='cuda:0')
    SparseMatrix('COO', 'scipy', 
                (3, 4))
    """
    def __new__(cls,  format:str, impl:str, *args, **kwargs):

        _initialize_if_needed()

        sparse_impl = _registered_sparse_formats[impl][format]

        return sparse_impl.constructor(*args, **kwargs)

def lookup_sparse_format(matrix: object) -> Optional[SparseMatrixFormat]:
    _initialize_if_needed()
    for sp_bkend in _registered_sparse_formats.values():
        for sp_fmt in sp_bkend.values():
            if sp_fmt.is_of_this_sparse_format(matrix):
                return sp_fmt
    return None
    
def is_sparse(matrix):
    return (lookup_sparse_format(matrix) is not None)

def get_sparse_matrix_impl(matrix):
    instance = lookup_sparse_format(matrix)
    assert instance is not None, 'The matrix is not a supported sparse matrix'
    return instance.impl

def get_sparse_matrix_format(matrix):
    instance = lookup_sparse_format(matrix)
    assert instance is not None, 'The matrix is not a supported sparse matrix'
    return instance.sparse_format

if __name__ == '__main__':
    print(SparseMatrix('COO', 'pytorch', 
                       [[0, 1, 1],[2, 0, 2]], [3, 4, 5], 
                       device='cuda:0'))
    print(SparseMatrix('COO', 'scipy', (3, 4)))



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Solvers for the optimization of the difference of convex functions.

Collection of DCA (d.c. algorithms) and related methods which make use of
structured optimization if the objective function can be written as a
difference of two convex functions.
"""

from __future__ import print_function, division, absolute_import

__all__ = ('dca', 'prox_dca', 'doubleprox_dc')


def dca(x, f, g, niter, callback=None):
    r"""Subgradient DCA of Tao and An.

    This algorithm solves a problem of the form ::

        min_x f(x) - g(x),

    where ``f`` and ``g`` are proper, convex and lower semicontinuous
    functions.

    Parameters
    ----------
    x : `LinearSpaceElement`
        Initial point, updated in-place.
    f : `Functional`
        Convex functional. Needs to implement ``f.convex_conj.gradient``.
    g : `Functional`
        Convex functional. Needs to implement ``g.gradient``.
    niter : int
        Number of iterations.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Notes
    -----
    The algorithm is described in Section 3 and in particular in Theorem 3 of
    `[TA1997] <http://journals.math.ac.vn/acta/pdf/9701289.pdf>`_. The problem

    .. math::
        \min f(x) - g(x)

    has the first-order optimality condition :math:`0 \in \partial f(x) -
    \partial g(x)`, i.e., aims at finding an :math:`x` so that there exists a
    common element

    .. math::
        y \in \partial f(x) \cap \partial g(x).

    The element :math:`y` can be seen as a solution of the Toland dual problem

    .. math::
        \min g^*(y) - f^*(y)

    and the iteration is given by

    .. math::
        y_n \in \partial g(x_n), \qquad x_{n+1} \in \partial f^*(y_n),

    for :math:`n\geq 0`. Here, a subgradient is found by evaluating the
    gradient method of the respective functionals.

    References
    ----------
    [TA1997] Tao, P D, and An, L T H. *Convex analysis approach to d.c.
    programming: Theory, algorithms and applications*. Acta Mathematica
    Vietnamica, 22.1 (1997), pp 289--355.

    See also
    --------
    prox_dca :
        Solver with a proximal step for ``f`` and a subgradient step for ``g``.
    doubleprox_dc :
        Solver with proximal steps for all the nonsmooth convex functionals
        and a gradient step for a smooth functional.
    """
    space = f.domain
    if g.domain != space:
        raise ValueError('`f.domain` and `g.domain` need to be equal, but '
                         '{} != {}'.format(space, g.domain))
    f_convex_conj = f.convex_conj
    for _ in range(niter):
        f_convex_conj.gradient(g.gradient(x), out=x)

        if callback is not None:
            callback(x)


def prox_dca(x, f, g, niter, gamma, callback=None):
    r"""Proximal DCA of Sun, Sampaio and Candido.

    This algorithm solves a problem of the form ::

        min_x f(x) - g(x)

    where ``f`` and ``g`` are two proper, convex and lower semicontinuous
    functions.

    Parameters
    ----------
    x : `LinearSpaceElement`
        Initial point, updated in-place.
    f : `Functional`
        Convex functional. Needs to implement ``f.proximal``.
    g : `Functional`
        Convex functional. Needs to implement ``g.gradient``.
    niter : int
        Number of iterations.
    gamma : positive float
        Stepsize in the primal updates.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Notes
    -----
    The algorithm was proposed as Algorithm 2.3 in
    `[SSC2003]
    <http://www.global-sci.org/jcm/readabs.php?vol=21&no=4&page=451&year=2003&ppage=462>`_.
    It solves the problem

    .. math ::
        \min f(x) - g(x)

    by using subgradients of :math:`g` and proximal points of :math:`f`.
    The iteration is given by

    .. math ::
        y_n \in \partial g(x_n), \qquad x_{n+1}
            = \mathrm{Prox}_{\gamma f}(x_n + \gamma y_n).

    In contrast to `dca`, `prox_dca` uses proximal steps with respect to the
    convex part ``f``. Both algorithms use subgradients of the concave part
    ``g``.

    References
    ----------
    [SSC2003] Sun, W, Sampaio R J B, and Candido M A B. *Proximal point
    algorithm for minimization of DC function*. Journal of Computational
    Mathematics, 21.4 (2003), pp 451--462.

    See also
    --------
    dca :
        Solver with subgradinet steps for all the functionals.
    doubleprox_dc :
        Solver with proximal steps for all the nonsmooth convex functionals
        and a gradient step for a smooth functional.
    """
    space = f.domain
    if g.domain != space:
        raise ValueError('`f.domain` and `g.domain` need to be equal, but '
                         '{} != {}'.format(space, g.domain))
    for _ in range(niter):
        f.proximal(gamma)(x.lincomb(1, x, gamma, g.gradient(x)), out=x)

        if callback is not None:
            callback(x)


def doubleprox_dc(x, y, f, phi, g, K, niter, gamma, mu, callback=None):
    r"""Double-proxmial gradient d.c. algorithm of Banert and Bot.

    This algorithm solves a problem of the form ::

        min_x f(x) + phi(x) - g(Kx).

    Parameters
    ----------
    x : `LinearSpaceElement`
        Initial primal guess, updated in-place.
    y : `LinearSpaceElement`
        Initial dual guess, updated in-place.
    f : `Functional`
        Convex functional. Needs to implement ``g.proximal``.
    phi : `Functional`
        Convex functional. Needs to implement ``phi.gradient``.
        Convergence can be guaranteed if the gradient is Lipschitz continuous.
    g : `Functional`
        Convex functional. Needs to implement ``h.convex_conj.proximal``.
    K : `Operator`
        Linear operator. Needs to implement ``K.adjoint``
    niter : int
        Number of iterations.
    gamma : positive float
        Stepsize in the primal updates.
    mu : positive float
        Stepsize in the dual updates.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Notes
    -----
    This algorithm is proposed in `[BB2016]
    <https://arxiv.org/abs/1610.06538>`_ and solves the d.c. problem

    .. math ::
        \min_x f(x) + \varphi(x) - g(Kx)

    together with its Toland dual

    .. math ::
        \min_y g^*(y) - (f + \varphi)^*(K^* y).

    The iterations are given by

    .. math ::
        x_{n+1} &= \mathrm{Prox}_{\gamma f} (x_n + \gamma (K^* y_n
                   - \nabla \varphi(x_n))), \\
        y_{n+1} &= \mathrm{Prox}_{\mu g^*} (y_n + \mu K x_{n+1}).

    To guarantee convergence, the parameter :math:`\gamma` must satisfy
    :math:`0 < \gamma < 2/L` where :math:`L` is the Lipschitz constant of
    :math:`\nabla \varphi`.

    References
    ----------
    [BB2016] Banert, S, and Bot, R I. *A general double-proximal gradient
    algorithm for d.c. programming*. arXiv:1610.06538 [math.OC] (2016).

    See also
    --------
    dca :
        Solver with subgradient steps for all the functionals.
    prox_dca :
        Solver with a proximal step for ``f`` and a subgradient step for ``g``.
    """
    primal_space = f.domain
    dual_space = g.domain

    if phi.domain != primal_space:
        raise ValueError('`f.domain` and `phi.domain` need to be equal, but '
                         '{} != {}'.format(primal_space, phi.domain))
    if K.domain != primal_space:
        raise ValueError('`f.domain` and `K.domain` need to be equal, but '
                         '{} != {}'.format(primal_space, K.domain))
    if K.range != dual_space:
        raise ValueError('`g.domain` and `K.range` need to be equal, but '
                         '{} != {}'.format(dual_space, K.range))

    g_convex_conj = g.convex_conj
    for _ in range(niter):
        f.proximal(gamma)(x.lincomb(1, x,
                                    gamma, K.adjoint(y) - phi.gradient(x)),
                          out=x)
        g_convex_conj.proximal(mu)(y.lincomb(1, y, mu, K(x)), out=y)

        if callback is not None:
            callback(x)


def doubleprox_dc_simple(x, y, f, phi, g, K, niter, gamma, mu):
    """Non-optimized version of ``doubleprox_dc``.
    This function is intended for debugging. It makes a lot of copies and
    performs no error checking.
    """
    for _ in range(niter):
        f.proximal(gamma)(x + gamma * K.adjoint(y) -
                          gamma * phi.gradient(x), out=x)
        g.convex_conj.proximal(mu)(y + mu * K(x), out=y)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import pytest
import numpy as np

import odl
from odl.core.discr.grid import RectGrid, uniform_grid, sparse_meshgrid
from odl.core.util.testutils import all_equal


# ---- RectGrid ---- #


def test_RectGrid_init():
    sorted1 = np.array([2, 3, 4, 5])
    sorted2 = np.array([-4, -2, 0, 2, 4])
    sorted3 = np.linspace(-1, 2, 50)
    scalar = 0.5

    # Just test if the code runs
    RectGrid(sorted1)
    RectGrid(sorted1, sorted2)
    RectGrid(sorted1, sorted1)
    RectGrid(sorted1, sorted2, sorted3)
    RectGrid(sorted2, scalar, sorted1)


def test_RectGrid_init_raise():
    # Check different error scenarios

    # Good input
    sorted1 = np.array([2, 3, 4, 5])
    sorted2 = np.array([-4, -2, 0, 2, 4])

    # Bad input
    unsorted = np.arange(4)
    unsorted[2] = -1
    with_dups = np.arange(4)
    with_dups[3] = 2
    unsorted_with_dups = unsorted.copy()
    unsorted_with_dups[3] = 0
    with_nan = np.arange(4, dtype=float)
    with_nan[3] = np.nan
    with_inf = np.arange(4, dtype=float)
    with_inf[3] = np.inf
    empty = np.arange(0)
    bad_shape = np.array([[1, 2], [3, 4]])

    with pytest.raises(ValueError):
        RectGrid(sorted1, unsorted, sorted2)

    with pytest.raises(ValueError):
        RectGrid(sorted1, with_dups, sorted2)

    with pytest.raises(ValueError):
        RectGrid(sorted1, unsorted_with_dups, sorted2)

    with pytest.raises(ValueError):
        RectGrid(sorted1, with_nan, sorted2)

    with pytest.raises(ValueError):
        RectGrid(sorted1, with_inf, sorted2)

    with pytest.raises(ValueError):
        RectGrid(sorted1, empty, sorted2)

    with pytest.raises(ValueError):
        RectGrid(sorted1, bad_shape)


def test_RectGrid_ndim():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])
    vec3 = np.linspace(-2, 2, 50)

    grid1 = RectGrid(vec1)
    grid2 = RectGrid(vec1, vec2, vec3)

    assert grid1.ndim == 1
    assert grid2.ndim == 3


def test_RectGrid_shape():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.linspace(-2, 2, 50)
    scalar = 0.5

    grid1 = RectGrid(vec1)
    grid2 = RectGrid(vec1, vec2)
    grid3 = RectGrid(scalar, vec2)

    assert grid1.shape == (4,)
    assert grid2.shape == (4, 50)
    assert grid3.shape == (1, 50)


def test_RectGrid_size():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.linspace(-2, 2, 50)
    scalar = 0.5

    grid1 = RectGrid(vec1)
    grid2 = RectGrid(vec1, vec2)
    grid3 = RectGrid(scalar, vec2)

    assert grid1.size == 4
    assert grid2.size == 200
    assert grid3.size == 50


def test_RectGrid_minpt_maxpt():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])
    vec3 = np.array([-1, 0])
    scalar = 0.5

    grid = RectGrid(vec1, vec2, vec3)
    assert all_equal(grid.min_pt, (2, -4, -1))
    assert all_equal(grid.max_pt, (5, 4, 0))

    grid = RectGrid(vec1, scalar, vec2, scalar)
    assert all_equal(grid.min_pt, (2, 0.5, -4, 0.5))
    assert all_equal(grid.max_pt, (5, 0.5, 4, 0.5))


def test_RectGrid_element():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])

    grid = RectGrid(vec1, vec2)
    some_pt = grid.element()
    assert some_pt in grid


def _test_eq(x, y):
    """Test equality of x and y."""
    assert x == y
    assert not x != y
    assert hash(x) == hash(y)


def _test_neq(x, y):
    """Test non-equality of x and y."""
    assert x != y
    assert not x == y
    assert hash(x) != hash(y)


def test_RectGrid_equals():
    """Test grid equality checks and hash."""
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])

    grid1 = RectGrid(vec1)
    grid2 = RectGrid(vec1, vec2)
    grid2_again = RectGrid(vec1, vec2)
    grid2_rev = RectGrid(vec2, vec1)

    _test_eq(grid1, grid1)
    _test_eq(grid2, grid2)
    _test_eq(grid2, grid2_again)

    _test_neq(grid1, grid2)
    _test_neq(grid2, grid2_rev)
    assert grid2 != (vec1, vec2)

    # Fuzzy check
    grid1 = RectGrid(vec1, vec2)
    grid2 = RectGrid(vec1 + (0.1, 0.05, 0, -0.1),
                     vec2 + (0.1, 0.05, 0, -0.1, -0.1))
    assert grid1.approx_equals(grid1, atol=0.0)
    assert grid1.approx_equals(grid2, atol=0.15)
    assert grid2.approx_equals(grid1, atol=0.15)

    grid2 = RectGrid(vec1 + (0.11, 0.05, 0, -0.1),
                     vec2 + (0.1, 0.05, 0, -0.1, -0.1))
    assert not grid1.approx_equals(grid2, atol=0.1)
    grid2 = RectGrid(vec1 + (0.1, 0.05, 0, -0.1),
                     vec2 + (0.1, 0.05, 0, -0.11, -0.1))
    assert not grid1.approx_equals(grid2, atol=0.1)


def test_RectGrid_contains():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])

    grid = RectGrid(vec1, vec2)

    point_list = []
    for x in vec1:
        for y in vec2:
            point_list.append((x, y))

    assert all(p in grid for p in point_list)

    assert not (0, 0) in grid
    assert (0, 0) not in grid
    assert (2, 0, 0) not in grid
    assert [None, object] not in grid

    # Fuzzy check
    assert grid.approx_contains((2.1, -2.1), atol=0.15)
    assert not grid.approx_contains((2.2, -2.1), atol=0.15)

    # 1d points
    grid = RectGrid(vec1)
    assert 3 in grid
    assert 7 not in grid


def test_RectGrid_is_subgrid():
    vec1 = np.array([2, 3, 4, 5])
    vec1_sup = np.array([2, 3, 4, 5, 6, 7])
    vec2 = np.array([-4, -2, 0, 2, 4])
    vec2_sup = np.array([-6, -4, -2, 0, 2, 4, 6])
    vec2_sub = np.array([-4, -2, 0, 2])
    scalar = 0.5

    grid = RectGrid(vec1, vec2)
    assert grid.is_subgrid(grid)

    sup_grid = RectGrid(vec1_sup, vec2_sup)
    assert grid.is_subgrid(sup_grid)
    assert not sup_grid.is_subgrid(grid)

    not_sup_grid = RectGrid(vec1_sup, vec2_sub)
    assert not grid.is_subgrid(not_sup_grid)
    assert not not_sup_grid.is_subgrid(grid)

    # Fuzzy check
    fuzzy_vec1_sup = vec1_sup + (0.1, 0.05, 0, -0.1, 0, 0.1)
    fuzzy_vec2_sup = vec2_sup + (0.1, 0.05, 0, -0.1, 0, 0.1, 0.05)
    fuzzy_sup_grid = RectGrid(fuzzy_vec1_sup, fuzzy_vec2_sup)
    assert grid.is_subgrid(fuzzy_sup_grid, atol=0.15)

    fuzzy_vec2_sup = vec2_sup + (0.1, 0.05, 0, -0.1, 0, 0.11, 0.05)
    fuzzy_sup_grid = RectGrid(fuzzy_vec1_sup, fuzzy_vec2_sup)
    assert not grid.is_subgrid(fuzzy_sup_grid, atol=0.1)

    # Changes in the non-overlapping part don't matter
    fuzzy_vec2_sup = vec2_sup + (0.1, 0.05, 0, -0.1, 0, 0.05, 0.11)
    fuzzy_sup_grid = RectGrid(fuzzy_vec1_sup, fuzzy_vec2_sup)
    assert grid.is_subgrid(fuzzy_sup_grid, atol=0.15)

    # With degenerate axis
    grid = RectGrid(vec1, scalar, vec2)
    sup_grid = RectGrid(vec1_sup, scalar, vec2_sup)
    assert grid.is_subgrid(sup_grid)

    fuzzy_sup_grid = RectGrid(vec1, scalar + 0.1, vec2)
    assert grid.is_subgrid(fuzzy_sup_grid, atol=0.15)


def test_RectGrid_insert():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])
    vec3 = np.array([-1, 0])
    scalar = 0.5

    grid = RectGrid(vec1, vec2)
    grid2 = RectGrid(scalar, vec3)

    # Test all positions
    ins_grid = grid.insert(0, grid2)
    assert ins_grid == RectGrid(scalar, vec3, vec1, vec2)

    ins_grid = grid.insert(1, grid2)
    assert ins_grid == RectGrid(vec1, scalar, vec3, vec2)

    ins_grid = grid.insert(2, grid2)
    assert ins_grid == RectGrid(vec1, vec2, scalar, vec3)

    ins_grid = grid.insert(-1, grid2)
    assert ins_grid == RectGrid(vec1, scalar, vec3, vec2)

    with pytest.raises(IndexError):
        grid.insert(3, grid2)

    with pytest.raises(IndexError):
        grid.insert(-4, grid2)


def test_RectGrid_points():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])
    scalar = 0.5

    # C ordering
    points = []
    for x1 in vec1:
        for x2 in vec2:
            points.append(np.array((x1, x2), dtype=float))

    grid = RectGrid(vec1, vec2)
    assert all_equal(points, grid.points())
    assert all_equal(points, grid.points(order='C'))
    assert all_equal(grid.min_pt, grid.points()[0])
    assert all_equal(grid.max_pt, grid.points()[-1])

    # F ordering
    points = []
    for x2 in vec2:
        for x1 in vec1:
            points.append(np.array((x1, x2), dtype=float))

    grid = RectGrid(vec1, vec2)
    assert all_equal(points, grid.points(order='F'))

    # Degenerate axis 1
    points = []
    for x1 in vec1:
        for x2 in vec2:
            points.append(np.array((scalar, x1, x2), dtype=float))

    grid = RectGrid(scalar, vec1, vec2)
    assert all_equal(points, grid.points())

    # Degenerate axis 2
    points = []
    for x1 in vec1:
        for x2 in vec2:
            points.append(np.array((x1, scalar, x2), dtype=float))

    grid = RectGrid(vec1, scalar, vec2)
    assert all_equal(points, grid.points())

    # Degenerate axis 3
    points = []
    for x1 in vec1:
        for x2 in vec2:
            points.append(np.array((x1, x2, scalar), dtype=float))

    grid = RectGrid(vec1, vec2, scalar)
    assert all_equal(points, grid.points())

    # Bad input
    with pytest.raises(ValueError):
        grid.points(order='A')


def test_RectGrid_corners():
    vec1 = np.array([2, 3, 4, 5])
    vec2 = np.array([-4, -2, 0, 2, 4])
    vec3 = np.array([-1, 0])
    scalar = 0.5

    minmax1 = (vec1[0], vec1[-1])
    minmax2 = (vec2[0], vec2[-1])
    minmax3 = (vec3[0], vec3[-1])

    # C ordering
    corners = []
    for x1 in minmax1:
        for x2 in minmax2:
            for x3 in minmax3:
                corners.append(np.array((x1, x2, x3), dtype=float))

    grid = RectGrid(vec1, vec2, vec3)
    assert all_equal(corners, grid.corners())
    assert all_equal(corners, grid.corners(order='C'))

    # minpt and maxpt should appear at the beginning and the end, resp.
    assert all_equal(grid.min_pt, grid.corners()[0])
    assert all_equal(grid.max_pt, grid.corners()[-1])

    # F ordering
    corners = []
    for x3 in minmax3:
        for x2 in minmax2:
            for x1 in minmax1:
                corners.append(np.array((x1, x2, x3), dtype=float))

    assert all_equal(corners, grid.corners(order='F'))

    # Degenerate axis 1
    corners = []
    for x2 in minmax2:
        for x3 in minmax3:
            corners.append(np.array((scalar, x2, x3), dtype=float))

    grid = RectGrid(scalar, vec2, vec3)
    assert all_equal(corners, grid.corners())

    # Degenerate axis 2
    corners = []
    for x1 in minmax1:
        for x3 in minmax3:
            corners.append(np.array((x1, scalar, x3), dtype=float))

    grid = RectGrid(vec1, scalar, vec3)
    assert all_equal(corners, grid.corners())

    # Degenerate axis 3
    corners = []
    for x1 in minmax1:
        for x2 in minmax2:
            corners.append(np.array((x1, x2, scalar), dtype=float))

    grid = RectGrid(vec1, vec2, scalar)
    assert all_equal(corners, grid.corners())

    # All degenerate
    corners = [(scalar, scalar)]
    grid = RectGrid(scalar, scalar)
    assert all_equal(corners, grid.corners())


def test_RectGrid_meshgrid():
    vec1 = (0, 1)
    vec2 = (-1, 0, 1)
    vec3 = (2, 3, 4, 5)

    # Sparse meshgrid
    mgx = np.array(vec1)[:, None, None]
    mgy = np.array(vec2)[None, :, None]
    mgz = np.array(vec3)[None, None, :]

    grid = RectGrid(vec1, vec2, vec3)
    xx, yy, zz = grid.meshgrid
    assert all_equal(mgx, xx)
    assert all_equal(mgy, yy)
    assert all_equal(mgz, zz)

    xx, yy, zz = grid.meshgrid
    assert all_equal(mgx, xx)
    assert all_equal(mgy, yy)
    assert all_equal(mgz, zz)


def test_RectGrid_getitem():
    vec1 = (0, 1, 2)
    vec2 = (-1, 0, 1)
    vec3 = (2, 3, 4, 5)
    vec4 = (1, 3)
    vec1_sub = (1,)
    vec2_sub = (-1,)
    vec3_sub = (3, 4)
    vec4_sub = (1,)

    grid = RectGrid(vec1, vec2, vec3, vec4)

    # Single indices yield points as an array
    assert all_equal(grid[1, 0, 1, 0], (1.0, -1.0, 3.0, 1.0))

    with pytest.raises(IndexError):
        grid[1, 0, 1, 0, 0]

    with pytest.raises(IndexError):
        grid[0, 3, 0, 0]

    # Slices return new RectGrid's
    assert grid == grid[...]

    sub_grid = RectGrid(vec1_sub, vec2_sub, vec3_sub, vec4_sub)
    assert grid[1, 0, 1:3, 0] == sub_grid
    assert grid[-2, :1, 1:3, :1] == sub_grid
    assert grid[1, 0, ..., 1:3, 0] == sub_grid

    sub_grid = RectGrid(vec1_sub, vec2, vec3, vec4)
    assert grid[1, :, :, :] == sub_grid
    assert grid[1, ...] == sub_grid
    assert grid[1] == sub_grid

    sub_grid = RectGrid(vec1, vec2, vec3, vec4_sub)
    assert grid[:, :, :, 0] == sub_grid
    assert grid[..., 0] == sub_grid

    sub_grid = RectGrid(vec1_sub, vec2, vec3, vec4_sub)
    assert grid[1, :, :, 0] == sub_grid
    assert grid[1, ..., 0] == sub_grid
    assert grid[1, :, :, ..., 0] == sub_grid

    # Fewer indices
    assert grid[0] == grid[0, :, :, :]
    assert grid[0, 1:] == grid[0, 1:, :, :]
    assert grid[0, 1:, :-1] == grid[0, 1:, :-1, :]

    # Indexing with lists
    sub_grid = RectGrid([0, 2], vec2, vec3, vec4)
    assert grid[[0, 2]] == sub_grid
    assert grid[[0, 1]] == grid[0:2, ...]

    # Two ellipses not allowed
    with pytest.raises(ValueError):
        grid[1, ..., ..., 0]

    # Too many indices
    with pytest.raises(IndexError):
        grid[1, 0, 1:2, 0, :]

    # Empty axes not allowed
    with pytest.raises(ValueError):
        grid[1, 0, None, 0]

    # One-dimensional grid
    grid = RectGrid(vec3)
    assert grid == grid[...]

    sub_grid = RectGrid(vec3_sub)
    assert grid[1:3] == sub_grid


def test_empty_grid():
    """Check if empty grids behave as expected and all methods work."""
    grid = RectGrid()

    assert grid.ndim == grid.size == len(grid) == 0
    assert grid.shape == ()

    assert grid.coord_vectors == ()
    assert grid.nondegen_byaxis == ()
    assert np.array_equal(grid.min_pt, [])
    assert np.array_equal(grid.max_pt, [])
    assert np.array_equal(grid.mid_pt, [])
    assert np.array_equal(grid.stride, [])
    assert np.array_equal(grid.extent, [])
    out = np.array([])
    grid.min(out=out)
    grid.max(out=out)

    assert grid.is_uniform
    assert grid.convex_hull() == odl.IntervalProd([], [])

    same = RectGrid()
    assert grid == same
    assert hash(grid) == hash(same)
    other = RectGrid([0, 2, 3])
    assert grid != other
    assert grid.is_subgrid(other)
    assert [] in grid
    assert 1.0 not in grid

    assert grid.insert(0, other) == other
    assert other.insert(0, grid) == other
    assert other.insert(1, grid) == other
    assert grid.squeeze() == grid
    assert np.array_equal(grid.points(), np.array([]).reshape((0, 0)))
    assert grid.corner_grid() == grid
    assert np.array_equal(grid.corners(), np.array([]).reshape((0, 0)))
    assert grid.meshgrid == ()

    assert grid[[]] == grid
    assert np.array_equal(np.asarray(grid), np.array([]).reshape((0, 0)))
    assert grid == uniform_grid([], [], ())
    repr(grid)


# ---- uniform_grid ---- #


def test_uniform_init():
    minpt = (0.75, 0, -5)
    maxpt = (1.25, 0, 1)
    shape = (2, 1, 3)

    # Check correct initialization of coord_vectors
    grid = uniform_grid(minpt, maxpt, shape)
    vec1 = (0.75, 1.25)
    vec2 = (0,)
    vec3 = (-5, -2, 1)
    assert all_equal(grid.coord_vectors, (vec1, vec2, vec3))


def test_uniform_init_raise():
    # Check different error scenarios
    minpt = (0.75, 0, -5)
    maxpt = (1.25, 0, 1)
    shape = (2, 1, 3)
    nonpos_shape1 = (2, 0, 3)
    nonpos_shape2 = (-2, 1, 3)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt, nonpos_shape1)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt, nonpos_shape2)

    minpt_with_nan = (0.75, 0, np.nan)
    minpt_with_inf = (0.75, 0, np.inf)
    maxpt_with_nan = (1.25, np.nan, 1)
    maxpt_with_inf = (1.25, np.inf, 1)
    shape_with_nan = (2, np.nan, 3)
    shape_with_inf = (2, np.inf, 3)

    with pytest.raises(ValueError):
        uniform_grid(minpt_with_nan, maxpt, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt_with_inf, maxpt, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt_with_nan, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt_with_inf, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt, shape_with_nan)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt, shape_with_inf)

    maxpt_smaller_minpt1 = (0.7, 0, 1)
    maxpt_smaller_minpt2 = (1.25, -1, 1)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt_smaller_minpt1, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt_smaller_minpt2, shape)

    too_short_minpt = (0.75, 0)
    too_long_minpt = (0.75, 0, -5, 2)
    too_short_maxpt = (0, 1)
    too_long_maxpt = (1.25, 0, 1, 25)
    too_short_shape = (2, 3)
    too_long_shape = (2, 1, 4, 3)
    bad_dim_shape = ((1, 2), (3, 4))

    with pytest.raises(ValueError):
        uniform_grid(too_short_minpt, maxpt, shape)

    with pytest.raises(ValueError):
        uniform_grid(too_long_minpt, maxpt, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, too_short_maxpt, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, too_long_maxpt, shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt, too_short_shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt, too_long_shape)

    maxpt_eq_minpt_at_shape_larger_than_1 = (0.75, 0, 1)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt_eq_minpt_at_shape_larger_than_1,
                     shape)

    with pytest.raises(ValueError):
        uniform_grid(minpt, maxpt, bad_dim_shape)


def test_uniform_mid_pt():
    minpt = (0.75, 0, -5)
    maxpt = (1.25, 0, 1)
    shape = (2, 1, 3)

    mid_pt = (1, 0, -2)

    grid = uniform_grid(minpt, maxpt, shape)
    assert all_equal(grid.mid_pt, mid_pt)


def test_uniform_stride():
    minpt = (0.75, 0, -5)
    maxpt = (1.25, 0, 1)
    shape = (2, 1, 3)

    stride = (0.5, 0, 3)

    grid = uniform_grid(minpt, maxpt, shape)
    assert all_equal(grid.stride, stride)


def test_uniform_is_subgrid():
    minpt = (0.75, 0, -5)
    maxpt = (1.25, 0, 1)
    shape = (2, 1, 5)

    # Optimized cases
    grid = uniform_grid(minpt, maxpt, shape)
    assert grid.is_subgrid(grid)

    smaller_shape = (1, 1, 5)
    not_sup_grid = uniform_grid(minpt, maxpt, smaller_shape)
    assert not grid.is_subgrid(not_sup_grid)

    larger_minpt = (0.85, 0, -4)
    not_sup_grid = uniform_grid(larger_minpt, maxpt, shape)
    assert not grid.is_subgrid(not_sup_grid)

    smaller_maxpt = (1.15, 0, 0)
    not_sup_grid = uniform_grid(minpt, smaller_maxpt, shape)
    assert not grid.is_subgrid(not_sup_grid)

    # Real checks
    minpt_sup1 = (-0.25, -2, -5)
    maxpt_sup1 = (1.25, 2, 1)
    shape_sup1 = (4, 3, 9)
    sup_grid = uniform_grid(minpt_sup1, maxpt_sup1, shape_sup1)
    assert grid.is_subgrid(sup_grid)
    assert not sup_grid.is_subgrid(grid)

    minpt_sup2 = (0.5, 0, -5)
    maxpt_sup2 = (1.5, 0, 1)
    shape_sup2 = (5, 1, 9)
    sup_grid = uniform_grid(minpt_sup2, maxpt_sup2, shape_sup2)
    assert grid.is_subgrid(sup_grid)
    assert not sup_grid.is_subgrid(grid)

    shape_not_sup1 = (4, 3, 10)
    not_sup_grid = uniform_grid(minpt_sup1, maxpt_sup1, shape_not_sup1)
    assert not grid.is_subgrid(not_sup_grid)
    assert not not_sup_grid.is_subgrid(grid)

    minpt_not_sup1 = (-0.25, -2.5, -5)
    not_sup_grid = uniform_grid(minpt_not_sup1, maxpt_sup1, shape_sup1)
    assert not grid.is_subgrid(not_sup_grid)
    assert not not_sup_grid.is_subgrid(grid)

    maxpt_not_sup1 = (1.35, 2.0001, 1)
    not_sup_grid = uniform_grid(minpt_sup1, maxpt_not_sup1, shape_sup1)
    assert not grid.is_subgrid(not_sup_grid)
    assert not not_sup_grid.is_subgrid(grid)

    # Should also work for RectGrid's
    vec1_sup = (0.75, 1, 1.25, 7)
    vec2_sup = (0,)
    vec3_sup = (-5, -3.5, -3, -2, -0.5, 0, 1, 9.5)

    tensor_sup_grid = RectGrid(vec1_sup, vec2_sup, vec3_sup)
    assert grid.is_subgrid(tensor_sup_grid)

    vec1_not_sup = (1, 1.25, 7)
    vec2_not_sup = (0,)
    vec3_not_sup = (-4, -2, 1)

    tensor_not_sup_grid = RectGrid(vec1_not_sup, vec2_not_sup, vec3_not_sup)
    assert not grid.is_subgrid(tensor_not_sup_grid)

    # Fuzzy check
    shape_sup = (4, 3, 9)

    minpt_fuzzy_sup1 = (-0.24, -2, -5.01)
    minpt_fuzzy_sup2 = (-0.24, -2, -5)
    maxpt_fuzzy_sup1 = (1.24, 2, 1)
    maxpt_fuzzy_sup2 = (1.25, 2, 1.01)

    fuzzy_sup_grid = uniform_grid(minpt_fuzzy_sup1, maxpt_fuzzy_sup1,
                                  shape_sup)
    assert grid.is_subgrid(fuzzy_sup_grid, atol=0.015)
    assert not grid.is_subgrid(fuzzy_sup_grid, atol=0.005)

    fuzzy_sup_grid = uniform_grid(minpt_fuzzy_sup2, maxpt_fuzzy_sup2,
                                  shape_sup)
    assert grid.is_subgrid(fuzzy_sup_grid, atol=0.015)
    assert not grid.is_subgrid(fuzzy_sup_grid, atol=0.005)


def test_uniform_insert():
    minpt = (0.75, 0, -5)
    maxpt = (1.25, 0, 1)
    shape = (2, 1, 3)

    minpt2 = (1, 1)
    maxpt2 = (3, 1)
    shape2 = (5, 1)

    grid = uniform_grid(minpt, maxpt, shape)
    grid2 = uniform_grid(minpt2, maxpt2, shape2)

    # Test all positions
    ins_grid = grid.insert(0, grid2)
    ins_minpt = minpt2 + minpt
    ins_maxpt = maxpt2 + maxpt
    ins_shape = shape2 + shape
    assert ins_grid == uniform_grid(ins_minpt, ins_maxpt, ins_shape)

    ins_grid = grid.insert(1, grid2)
    ins_minpt = minpt[:1] + minpt2 + minpt[1:]
    ins_maxpt = maxpt[:1] + maxpt2 + maxpt[1:]
    ins_shape = shape[:1] + shape2 + shape[1:]
    assert ins_grid == uniform_grid(ins_minpt, ins_maxpt, ins_shape)

    ins_grid = grid.insert(2, grid2)
    ins_minpt = minpt[:2] + minpt2 + minpt[2:]
    ins_maxpt = maxpt[:2] + maxpt2 + maxpt[2:]
    ins_shape = shape[:2] + shape2 + shape[2:]
    assert ins_grid == uniform_grid(ins_minpt, ins_maxpt, ins_shape)
    ins_grid = grid.insert(-1, grid2)
    assert ins_grid == uniform_grid(ins_minpt, ins_maxpt, ins_shape)

    ins_grid = grid.insert(3, grid2)
    ins_minpt = minpt + minpt2
    ins_maxpt = maxpt + maxpt2
    ins_shape = shape + shape2
    assert ins_grid == uniform_grid(ins_minpt, ins_maxpt, ins_shape)

    # Insert a RectGrid
    vec = [-1, 0, 3]
    tgrid = RectGrid(vec)
    ins_tgrid = grid.insert(3, tgrid)
    assert isinstance(ins_tgrid, RectGrid)
    assert ins_tgrid == RectGrid(*(grid.coord_vectors + (vec,)))

    with pytest.raises(IndexError):
        grid.insert(4, grid2)

    with pytest.raises(IndexError):
        grid.insert(-5, grid2)

    with pytest.raises(TypeError):
        grid.insert(0, [1, 2])


def test_uniform_getitem():
    minpt = (0.75, 0, -5, 4)
    maxpt = (1.25, 0, 1, 13)
    shape = (2, 1, 5, 4)

    grid = uniform_grid(minpt, maxpt, shape)

    # Single indices yield points as an array
    indices = [1, 0, 1, 1]
    values = [vec[i] for i, vec in zip(indices, grid.coord_vectors)]
    assert all_equal(grid[1, 0, 1, 1], values)

    indices = [0, 0, 4, 3]
    values = [vec[i] for i, vec in zip(indices, grid.coord_vectors)]
    assert all_equal(grid[0, 0, 4, 3], values)

    with pytest.raises(IndexError):
        grid[1, 0, 1, 2, 0]

    with pytest.raises(IndexError):
        grid[1, 1, 6, 2]

    with pytest.raises(IndexError):
        grid[1, 0, 4, 6]

    # Slices return uniform grids
    assert grid == grid[...]

    # Use RectGrid implementation as reference here
    tensor_grid = RectGrid(*grid.coord_vectors)

    test_slice = np.s_[1, :, ::2, ::3]
    assert all_equal(grid[test_slice].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[1:2, :, ::2, ::3].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[1:2, :, ::2, ..., ::3].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)

    test_slice = np.s_[0:1, :, :, 2:4]
    assert all_equal(grid[test_slice].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[:1, :, :, 2:].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[:-1, ..., 2:].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)

    test_slice = np.s_[:, 0, :, :]
    assert all_equal(grid[test_slice].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[:, 0, ...].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[0:2, :, ...].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[...].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)

    test_slice = np.s_[:, :, 0::2, :]
    assert all_equal(grid[test_slice].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[..., 0::2, :].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)

    test_slice = np.s_[..., 1, :]
    assert all_equal(grid[test_slice].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)
    assert all_equal(grid[:, :, 1, :].coord_vectors,
                     tensor_grid[test_slice].coord_vectors)

    # Fewer indices
    assert grid[1:] == grid[1:, :, :, :]
    assert grid[1:, 0] == grid[1:, 0, :, :]
    assert grid[1:, 0, :-1] == grid[1:, 0, :-1, :]

    # Two ellipses not allowed
    with pytest.raises(ValueError):
        grid[1, ..., ..., 0]

    # Too many axes
    with pytest.raises(IndexError):
        grid[1, 0, 1:2, 0, :]

    # New axes not supported
    with pytest.raises(ValueError):
        grid[1, 0, None, 1, 0]

    # Empty axes not allowed
    with pytest.raises(ValueError):
        grid[1, 0, 0:0, 1]
    with pytest.raises(ValueError):
        grid[1, 1:, 0, 1]

    # One-dimensional grid
    grid = uniform_grid(1, 5, 5)
    assert grid == grid[...]

    sub_grid = uniform_grid(1, 5, 3)
    assert grid[::2], sub_grid


# ---- Utilities ---- #

def test_sparse_meshgrid():

    # One array only
    x = np.zeros(2)
    true_mg = (x,)
    assert all_equal(sparse_meshgrid(x), true_mg)

    x = np.zeros((2, 2))
    true_mg = (x,)
    assert all_equal(sparse_meshgrid(x), true_mg)

    # Two arrays
    x, y = np.zeros(2), np.zeros(3)
    true_mg = (x[:, None], y[None, :])
    mg = sparse_meshgrid(x, y)
    assert all_equal(mg, true_mg)
    assert all(vec.flags.c_contiguous for vec in mg)

    # Array-like input
    x, y = [1, 2, 3], [4, 5, 6]
    true_mg = (np.array(x)[:, None], np.array(y)[None, :])
    mg = sparse_meshgrid(x, y)
    assert all_equal(mg, true_mg)
    assert all(vec.flags.c_contiguous for vec in mg)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Sparse implementations of n-dimensional sampling grids.

Sampling grids are collections of points in an n-dimensional coordinate
space with a certain structure which is exploited to minimize storage.
"""

from __future__ import print_function, division, absolute_import
import numpy as np

from odl.core.set import Set, IntervalProd
from odl.core.util import (
    normalized_index_expression, normalized_scalar_param_list, safe_int_conv,
    array_str, signature_string, indent, npy_printoptions)


__all__ = (
    'sparse_meshgrid',
    'RectGrid',
    'uniform_grid',
    'uniform_grid_fromintv',
)


def sparse_meshgrid(*x):
    """Make a sparse `meshgrid` by adding empty dimensions.

    Parameters
    ----------
    x1,...,xN : `array-like`
        Input arrays to turn into sparse meshgrid vectors.

    Returns
    -------
    meshgrid : tuple of `numpy.ndarray`'s
        Sparse coordinate vectors representing an N-dimensional grid.

    See Also
    --------
    numpy.meshgrid : dense or sparse meshgrids

    Examples
    --------
    >>> x, y = [0, 1], [2, 3, 4]
    >>> mesh = sparse_meshgrid(x, y)
    >>> np.sum(xi for xi in mesh).ravel()  # first axis slowest
    array([2, 3, 4, 3, 4, 5])
    """
    n = len(x)
    mesh = []
    for ax, xi in enumerate(x):
        xi = np.asarray(xi)
        slc = [None] * n
        slc[ax] = slice(None)
        mesh.append(np.ascontiguousarray(xi[tuple(slc)]))

    return tuple(mesh)


class RectGrid(Set):

    """An n-dimensional rectilinear grid.

    A rectilinear grid is the set of points defined by all possible
    combination of coordinates taken from fixed coordinate vectors.

    The storage need for a rectilinear grid is only the sum of the lengths
    of the coordinate vectors, while the total number of points is
    the product of these lengths. This class makes use of that
    sparse storage scheme.

    See ``Notes`` for details.
    """

    def __init__(self, *coord_vectors):
        r"""Initialize a new instance.

        Parameters
        ----------
        vec1,...,vecN : `array-like`
            The coordinate vectors defining the grid points. They must
            be sorted in ascending order and may not contain
            duplicates. Empty vectors are not allowed.

        Examples
        --------
        >>> g = RectGrid([1, 2, 5], [-2, 1.5, 2])
        >>> g
        RectGrid(
            [ 1.,  2.,  5.],
            [-2. ,  1.5,  2. ]
        )
        >>> g.ndim  # number of axes
        2
        >>> g.shape  # points per axis
        (3, 3)
        >>> g.size  # total number of points
        9

        Grid points can be extracted with index notation (NOTE: This is
        slow, do not loop over the grid using indices!):

        >>> g = RectGrid([-1, 0, 3], [2, 4, 5], [5], [2, 4, 7])
        >>> g[0, 0, 0, 0]
        array([-1.,  2.,  5.,  2.])

        Slices and ellipsis are also supported:

        >>> g[:, 0, 0, 0]
        RectGrid(
            [-1.,  0.,  3.],
            [ 2.],
            [ 5.],
            [ 2.]
        )
        >>> g[0, ..., 1:]
        RectGrid(
            [-1.],
            [ 2.,  4.,  5.],
            [ 5.],
            [ 4.,  7.]
        )

        Notes
        -----
        In 2 dimensions, for example, given two coordinate vectors

        .. math::
            v_1 = (-1, 0, 2),\ v_2 = (0, 1)

        the corresponding rectilinear grid :math:`G` is the set of all
        2d points whose first component is from :math:`v_1` and the
        second component from :math:`v_2`:

        .. math::
            G = \{(-1, 0), (-1, 1), (0, 0), (0, 1), (2, 0), (2, 1)\}

        Here is a graphical representation::

               :    :        :
               :    :        :
            1 -x----x--------x-...
               |    |        |
            0 -x----x--------x-...
               |    |        |
              -1    0        2

        Apparently, this structure can represent grids with arbitrary step
        sizes in each axis.

        Note that the above ordering of points is the standard ``'C'``
        ordering where the first axis (:math:`v_1`) varies slowest.
        Ordering is only relevant when the point array is actually created;
        the grid itself is independent of this ordering.
        """
        super(RectGrid, self).__init__()

        vecs = tuple(np.atleast_1d(vec).astype('float64')
                     for vec in coord_vectors)
        for i, vec in enumerate(vecs):

            if len(vec) == 0:
                raise ValueError('vector {} has zero length'
                                 ''.format(i + 1))

            if not np.all(np.isfinite(vec)):
                raise ValueError('vector {} contains invalid entries'
                                 ''.format(i + 1))

            if vec.ndim != 1:
                raise ValueError('vector {} has {} dimensions instead of 1'
                                 ''.format(i + 1, vec.ndim))

            sorted_vec = np.sort(vec)
            if np.any(vec != sorted_vec):
                raise ValueError('vector {} not sorted'
                                 ''.format(i + 1))

            if np.any(np.diff(vec) == 0):
                raise ValueError('vector {} contains duplicates'
                                 ''.format(i + 1))

        # Lazily evaluates strides when needed but stores the result
        self.__stride = None

        self.__coord_vectors = vecs

        # Non-degenerate axes
        self.__nondegen_byaxis = tuple(len(v) > 1 for v in self.coord_vectors)

        # Uniformity, setting True in degenerate axes
        diffs = [np.diff(v) for v in self.coord_vectors]
        self.__is_uniform_byaxis = tuple(
            (diff.size == 0) or np.allclose(diff, diff[0])
            for diff in diffs)

    # Attributes
    @property
    def coord_vectors(self):
        """Coordinate vectors of the grid.

        Returns
        -------
        coord_vectors : tuple of `numpy.ndarray`'s

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1, 0, 2])
        >>> x, y = g.coord_vectors
        >>> x
        array([ 0.,  1.])
        >>> y
        array([-1.,  0.,  2.])

        See Also
        --------
        meshgrid : Same result but with nd arrays
        """
        return self.__coord_vectors

    @property
    def ndim(self):
        """Number of dimensions of the grid."""
        try:
            return self.__ndim
        except AttributeError:
            ndim = len(self.coord_vectors)
            self.__ndim = ndim
            return ndim

    @property
    def shape(self):
        """Number of grid points per axis."""
        try:
            return self.__shape
        except AttributeError:
            shape = tuple(len(vec) for vec in self.coord_vectors)
            self.__shape = shape
            return shape

    @property
    def size(self):
        """Total number of grid points."""
        # Since np.prod(()) == 1.0 we need to handle that by ourselves
        return (0 if self.shape == () else
                int(np.prod(self.shape, dtype='int64')))

    def __len__(self):
        """Return ``len(self)``.

        The length along the first dimension.

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1, 0, 2], [4, 5, 6])
        >>> len(g)
        2

        See Also
        --------
        size : The total number of elements.
        """
        return 0 if self.shape == () else self.shape[0]

    @property
    def min_pt(self):
        """Vector containing the minimal grid coordinates per axis.

        Examples
        --------
        >>> g = RectGrid([1, 2, 5], [-2, 1.5, 2])
        >>> g.min_pt
        array([ 1., -2.])
        """
        return np.array([vec[0] for vec in self.coord_vectors])

    @property
    def max_pt(self):
        """Vector containing the maximal grid coordinates per axis.

        Examples
        --------
        >>> g = RectGrid([1, 2, 5], [-2, 1.5, 2])
        >>> g.max_pt
        array([ 5.,  2.])
        """
        return np.array([vec[-1] for vec in self.coord_vectors])

    @property
    def nondegen_byaxis(self):
        """Boolean array with ``True`` entries for non-degenerate axes.

        Examples
        --------
        >>> g = uniform_grid([0, 0], [1, 1], (5, 1))
        >>> g.nondegen_byaxis
        (True, False)
        """
        return self.__nondegen_byaxis

    @property
    def is_uniform_byaxis(self):
        """Boolean tuple showing uniformity of this grid per axis."""
        return self.__is_uniform_byaxis

    @property
    def is_uniform(self):
        """``True`` if this grid is uniform in all axes, else ``False``."""
        return all(self.is_uniform_byaxis)

    # min, max and extent are for set duck-typing
    def min(self, **kwargs):
        """Return `min_pt`.

        Parameters
        ----------
        kwargs
            For duck-typing with `numpy.amin`

        See Also
        --------
        max
        odl.core.set.domain.IntervalProd.min

        Examples
        --------
        >>> g = RectGrid([1, 2, 5], [-2, 1.5, 2])
        >>> g.min()
        array([ 1., -2.])

        Also works with Numpy:

        >>> np.min(g)
        array([ 1., -2.])
        """
        out = kwargs.get('out', None)
        if out is not None:
            out[:] = self.min_pt
            return out
        else:
            return self.min_pt

    def max(self, **kwargs):
        """Return `max_pt`.

        Parameters
        ----------
        kwargs
            For duck-typing with `numpy.amax`

        See Also
        --------
        min
        odl.core.set.domain.IntervalProd.max

        Examples
        --------
        >>> g = RectGrid([1, 2, 5], [-2, 1.5, 2])
        >>> g.max()
        array([ 5.,  2.])

        Also works with Numpy:

        >>> np.max(g)
        array([ 5.,  2.])
        """
        out = kwargs.get('out', None)
        if out is not None:
            out[:] = self.max_pt
            return out
        else:
            return self.max_pt

    @property
    def mid_pt(self):
        """Midpoint of the grid, not necessarily a grid point.

        Examples
        --------
        >>> rg = uniform_grid([-1.5, -1], [-0.5, 3], (2, 3))
        >>> rg.mid_pt
        array([-1.,  1.])
        """
        return (self.max_pt + self.min_pt) / 2

    @property
    def stride(self):
        """Step per axis between neighboring points of a uniform grid.

        If the grid contains axes that are not uniform, ``stride`` has
        a ``NaN`` entry.

        For degenerate (length 1) axes, ``stride`` has value ``0.0``.

        Returns
        -------
        stride : numpy.array
            Array of dtype ``float`` and length `ndim`.

        Examples
        --------
        >>> rg = uniform_grid([-1.5, -1], [-0.5, 3], (2, 3))
        >>> rg.stride
        array([ 1.,  2.])

        NaN returned for non-uniform dimension:

        >>> g = RectGrid([0, 1, 2], [0, 1, 4])
        >>> g.stride
        array([ 1.,  nan])

        0.0 returned for degenerate dimension:

        >>> g = RectGrid([0, 1, 2], [0])
        >>> g.stride
        array([ 1.,  0.])
        """
        # Cache for efficiency instead of re-computing
        if self.__stride is None:
            strd = []
            for i in range(self.ndim):
                if not self.is_uniform_byaxis[i]:
                    strd.append(float('nan'))
                elif self.nondegen_byaxis[i]:
                    strd.append(self.extent[i] / (self.shape[i] - 1.0))
                else:
                    strd.append(0.0)
            self.__stride = np.array(strd)

        return self.__stride.copy()

    @property
    def extent(self):
        """Return the edge lengths of this grid's minimal bounding box.

        Examples
        --------
        >>> g = RectGrid([1, 2, 5], [-2, 1.5, 2])
        >>> g.extent
        array([ 4.,  4.])
        """
        return self.max_pt - self.min_pt

    def convex_hull(self):
        """Return the smallest `IntervalProd` containing this grid.

        The convex hull of a set is the union of all line segments
        between points in the set. For a rectilinear grid, it is the
        interval product given by the extremal coordinates.

        Returns
        -------
        convex_hull : `IntervalProd`
            Interval product defined by the minimum and maximum points
            of the grid.

        Examples
        --------
        >>> g = RectGrid([-1, 0, 3], [2, 4], [5], [2, 4, 7])
        >>> g.convex_hull()
        IntervalProd([-1.,  2.,  5.,  2.], [ 3.,  4.,  5.,  7.])
        """
        return IntervalProd(self.min(), self.max())

    def element(self):
        """An arbitrary element, the minimum coordinates."""
        return self.min_pt

    def approx_equals(self, other, atol):
        """Test if this grid is equal to another grid.

        Parameters
        ----------
        other :
            Object to be tested
        atol : float
            Allow deviations up to this number in absolute value
            per vector entry.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `RectGrid` instance with all
            coordinate vectors equal (up to the given tolerance), to
            the ones of this grid, ``False`` otherwise.

        Examples
        --------
        >>> g1 = RectGrid([0, 1], [-1, 0, 2])
        >>> g2 = RectGrid([-0.1, 1.1], [-1, 0.1, 2])
        >>> g1.approx_equals(g2, atol=0)
        False
        >>> g1.approx_equals(g2, atol=0.15)
        True
        """
        if other is self:
            return True

        return (type(other) is type(self) and
                self.ndim == other.ndim and
                self.shape == other.shape and
                all(np.allclose(vec_s, vec_o, atol=atol, rtol=0.0)
                    for (vec_s, vec_o) in zip(self.coord_vectors,
                                              other.coord_vectors)))

    def __eq__(self, other):
        """Return ``self == other``.
        """
        # Implemented separately for performance reasons
        if other is self:
            return True

        return (type(other) is type(self) and
                self.shape == other.shape and
                all(np.array_equal(vec_s, vec_o)
                    for (vec_s, vec_o) in zip(self.coord_vectors,
                                              other.coord_vectors)))

    def __hash__(self):
        """Return ``hash(self)``."""
        # TODO: update with #841
        coord_vec_str = tuple(cv.tobytes() for cv in self.coord_vectors)
        return hash((type(self), coord_vec_str))

    def approx_contains(self, other, atol):
        """Test if ``other`` belongs to this grid up to a tolerance.

        Parameters
        ----------
        other : `array-like` or float
            The object to test for membership in this grid
        atol : float
            Allow deviations up to this number in absolute value
            per vector entry.

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1, 0, 2])
        >>> g.approx_contains([0, 0], atol=0.0)
        True
        >>> [0, 0] in g  # equivalent
        True
        >>> g.approx_contains([0.1, -0.1], atol=0.0)
        False
        >>> g.approx_contains([0.1, -0.1], atol=0.15)
        True
        """
        other = np.atleast_1d(other)
        return (other.shape == (self.ndim,) and
                all(np.any(np.isclose(vector, coord, atol=atol, rtol=0.0))
                    for vector, coord in zip(self.coord_vectors, other)))

    def __contains__(self, other):
        """Return ``other in self``."""
        other = np.atleast_1d(other)
        if other.dtype == np.dtype(object):
            return False
        return (other.shape == (self.ndim,) and
                all(coord in vector
                    for vector, coord in zip(self.coord_vectors, other)))

    def is_subgrid(self, other, atol=0.0):
        """Return ``True`` if this grid is a subgrid of ``other``.

        Parameters
        ----------
        other :  `RectGrid`
            The other grid which is supposed to contain this grid
        atol : float, optional
            Allow deviations up to this number in absolute value
            per coordinate vector entry.

        Returns
        -------
        is_subgrid : bool
            ``True`` if all coordinate vectors of ``self`` are within
            absolute distance ``atol`` of the other grid, else ``False``.

        Examples
        --------
        >>> rg = uniform_grid([-2, -2], [0, 4], (3, 4))
        >>> rg.coord_vectors
        (array([-2., -1.,  0.]), array([-2.,  0.,  2.,  4.]))
        >>> rg_sub = uniform_grid([-1, 2], [0, 4], (2, 2))
        >>> rg_sub.coord_vectors
        (array([-1.,  0.]), array([ 2.,  4.]))
        >>> rg_sub.is_subgrid(rg)
        True

        Fuzzy check is also possible. Note that the tolerance still
        applies to the coordinate vectors.

        >>> rg_sub = uniform_grid([-1.015, 2], [0, 3.99], (2, 2))
        >>> rg_sub.is_subgrid(rg, atol=0.01)
        False
        >>> rg_sub.is_subgrid(rg, atol=0.02)
        True
        """
        # Optimization for some common cases
        if other is self:
            return True
        if not isinstance(other, RectGrid):
            return False
        if not all(self.shape[i] <= other.shape[i] and
                   self.min_pt[i] >= other.min_pt[i] - atol and
                   self.max_pt[i] <= other.max_pt[i] + atol
                   for i in range(self.ndim)):
            return False
        if self.size == 0:
            return True

        if self.is_uniform and other.is_uniform:
            # For uniform grids, it suffices to show that min_pt, max_pt
            # and g[1,...,1] are contained in the other grid. For axes
            # with less than 2 points, this reduces to min_pt and max_pt,
            # and the corresponding indices in the other check point are
            # set to 0.
            minmax_contained = (
                other.approx_contains(self.min_pt, atol=atol) and
                other.approx_contains(self.max_pt, atol=atol))
            check_idx = np.zeros(self.ndim, dtype=int)
            check_idx[np.array(self.shape) >= 3] = 1
            checkpt_contained = other.approx_contains(self[tuple(check_idx)],
                                                      atol=atol)
            return minmax_contained and checkpt_contained

        else:
            # Array version of the fuzzy subgrid test, about 3 times faster
            # than the loop version.
            for vec_o, vec_s in zip(other.coord_vectors, self.coord_vectors):
                # Create array of differences of all entries in vec_o and
                # vec_s. If there is no almost zero entry in each row,
                # return False.
                vec_o_mg, vec_s_mg = sparse_meshgrid(vec_o, vec_s)
                if not np.all(np.any(np.isclose(vec_s_mg, vec_o_mg, atol=atol),
                                     axis=0)):
                    return False
            return True

    def insert(self, index, *grids):
        """Return a copy with ``grids`` inserted before ``index``.

        The given grids are inserted (as a block) into ``self``, yielding
        a new grid whose number of dimensions is the sum of the numbers of
        dimensions of all involved grids.
        Note that no changes are made in-place.

        Parameters
        ----------
        index : int
            The index of the dimension before which ``grids`` are to
            be inserted. Negative indices count backwards from
            ``self.ndim``.
        grid1, ..., gridN :  `RectGrid`
            The grids to be inserted into ``self``.

        Returns
        -------
        newgrid : `RectGrid`
            The enlarged grid.

        Examples
        --------
        >>> g1 = RectGrid([0, 1], [-1, 0, 2])
        >>> g2 = RectGrid([1], [-6, 15])
        >>> g1.insert(1, g2)
        RectGrid(
            [ 0.,  1.],
            [ 1.],
            [ -6.,  15.],
            [-1.,  0.,  2.]
        )
        >>> g1.insert(1, g2, g2)
        RectGrid(
            [ 0.,  1.],
            [ 1.],
            [ -6.,  15.],
            [ 1.],
            [ -6.,  15.],
            [-1.,  0.,  2.]
        )

        See Also
        --------
        append
        """
        index, index_in = safe_int_conv(index), index
        if not -self.ndim <= index <= self.ndim:
            raise IndexError('index {0} outside the valid range -{1} ... {1}'
                             ''.format(index_in, self.ndim))
        if index < 0:
            index += self.ndim

        if len(grids) == 0:
            # Copy of `self`
            return RectGrid(*self.coord_vectors)
        elif len(grids) == 1:
            # Insert single grid
            grid = grids[0]
            if not isinstance(grid, RectGrid):
                raise TypeError('{!r} is not a `RectGrid` instance'
                                ''.format(grid))
            new_vecs = (self.coord_vectors[:index] + grid.coord_vectors +
                        self.coord_vectors[index:])
            return RectGrid(*new_vecs)
        else:
            # Recursively insert first grid and the remaining into the result
            return self.insert(index, grids[0]).insert(
                index + grids[0].ndim, *(grids[1:]))

    def append(self, *grids):
        """Insert ``grids`` at the end as a block.

        Parameters
        ----------
        grid1, ..., gridN :  `RectGrid`
            The grids to be appended to ``self``.

        Returns
        -------
        newgrid : `RectGrid`
            The enlarged grid.

        Examples
        --------
        >>> g1 = RectGrid([0, 1], [-1, 0, 2])
        >>> g2 = RectGrid([1], [-6, 15])
        >>> g1.append(g2)
        RectGrid(
            [ 0.,  1.],
            [-1.,  0.,  2.],
            [ 1.],
            [ -6.,  15.]
        )
        >>> g1.append(g2, g2)
        RectGrid(
            [ 0.,  1.],
            [-1.,  0.,  2.],
            [ 1.],
            [ -6.,  15.],
            [ 1.],
            [ -6.,  15.]
        )

        See Also
        --------
        insert
        """
        return self.insert(self.ndim, *grids)

    def squeeze(self, axis=None):
        """Return the grid with removed degenerate (length 1) dimensions.

        Parameters
        ----------
        axis : None or index expression, optional
            Subset of the axes to squeeze. Default: All axes.

        Returns
        -------
        squeezed : `RectGrid`
            Squeezed grid.

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1], [-1, 0, 2])
        >>> g.squeeze()
        RectGrid(
            [ 0.,  1.],
            [-1.,  0.,  2.]
        )
        """
        if axis is None:
            rng = range(self.ndim)
        else:
            rng = list(np.atleast_1d(np.arange(self.ndim)[axis]))

        new_indcs = [i for i in range(self.ndim)
                     if i not in rng or self.nondegen_byaxis[i]]
        coord_vecs = [self.coord_vectors[axis] for axis in new_indcs]
        return RectGrid(*coord_vecs)

    def points(self, order='C'):
        """All grid points in a single array.

        Parameters
        ----------
        order : {'C', 'F'}, optional
            Axis ordering in the resulting point array.

        Returns
        -------
        points : `numpy.ndarray`
            The shape of the array is ``size x ndim``, i.e. the points
            are stored as rows.

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1, 0, 2])
        >>> g.points()
        array([[ 0., -1.],
               [ 0.,  0.],
               [ 0.,  2.],
               [ 1., -1.],
               [ 1.,  0.],
               [ 1.,  2.]])
        >>> g.points(order='F')
        array([[ 0., -1.],
               [ 1., -1.],
               [ 0.,  0.],
               [ 1.,  0.],
               [ 0.,  2.],
               [ 1.,  2.]])
        """
        if str(order).upper() not in ('C', 'F'):
            raise ValueError('order {!r} not recognized'.format(order))
        else:
            order = str(order).upper()

        axes = range(self.ndim) if order == 'C' else reversed(range(self.ndim))
        shape = self.shape if order == 'C' else tuple(reversed(self.shape))
        point_arr = np.empty((self.size, self.ndim))

        for i, axis in enumerate(axes):
            view = point_arr[:, axis].reshape(shape)
            coord_shape = (1,) * i + (-1,) + (1,) * (self.ndim - i - 1)
            view[:] = self.coord_vectors[axis].reshape(coord_shape)

        return point_arr

    def corner_grid(self):
        """Return a grid with only the corner points.

        Returns
        -------
        cgrid : `RectGrid`
            Grid with size 2 in non-degenerate dimensions and 1
            in degenerate ones

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1, 0, 2])
        >>> g.corner_grid()
        uniform_grid([ 0., -1.], [ 1.,  2.], (2, 2))
        """
        minmax_vecs = []
        for axis in range(self.ndim):
            if self.shape[axis] == 1:
                minmax_vecs.append(self.coord_vectors[axis][0])
            else:
                minmax_vecs.append((self.coord_vectors[axis][0],
                                    self.coord_vectors[axis][-1]))

        return RectGrid(*minmax_vecs)

    def corners(self, order='C'):
        """Corner points of the grid in a single array.

        Parameters
        ----------
        order : {'C', 'F'}, optional
            Axis ordering in the resulting point array

        Returns
        -------
        corners : `numpy.ndarray`
            The size of the array is 2^m x ndim, where m is the number
            of non-degenerate axes, i.e. the corners are stored as rows.

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1, 0, 2])
        >>> g.corners()
        array([[ 0., -1.],
               [ 0.,  2.],
               [ 1., -1.],
               [ 1.,  2.]])
        >>> g.corners(order='F')
        array([[ 0., -1.],
               [ 1., -1.],
               [ 0.,  2.],
               [ 1.,  2.]])
        """
        return self.corner_grid().points(order=order)

    @property
    def meshgrid(self):
        """A grid suitable for function evaluation.

        Returns
        -------
        meshgrid : tuple of `numpy.ndarray`'s
            Function evaluation grid with ``ndim`` axes

        See Also
        --------
        numpy.meshgrid
            Coordinate matrices from coordinate vectors.
            We use ``indexing='ij'`` and ``copy=True``

        Examples
        --------
        >>> g = RectGrid([0, 1], [-1, 0, 2])
        >>> x, y = g.meshgrid
        >>> x
        array([[ 0.],
               [ 1.]])
        >>> y
        array([[-1.,  0.,  2.]])

        Easy function evaluation via broadcasting:

        >>> x ** 2 - y ** 2
        array([[-1.,  0., -4.],
               [ 0.,  1., -3.]])
        """
        return sparse_meshgrid(*self.coord_vectors)

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        Parameters
        ----------
        indices : index expression
            Object determining which parts of the grid to extract.
            ``None`` (new axis) and empty axes are not supported.

        Examples
        --------
        Indexing with integers along all axes produces an array (a point):

        >>> g = RectGrid([-1, 0, 3], [2, 4, 5], [5], [2, 4, 7])
        >>> g[0, 0, 0, 0]
        array([-1.,  2.,  5.,  2.])

        Otherwise, a new RectGrid is returned:

        >>> g[:, 0, 0, 0]
        RectGrid(
            [-1.,  0.,  3.],
            [ 2.],
            [ 5.],
            [ 2.]
        )
        >>> g[0, ..., 1:]
        RectGrid(
            [-1.],
            [ 2.,  4.,  5.],
            [ 5.],
            [ 4.,  7.]
        )
        >>> g[::2, ..., ::2]
        RectGrid(
            [-1.,  3.],
            [ 2.,  4.,  5.],
            [ 5.],
            [ 2.,  7.]
        )

        Too few indices are filled up with an ellipsis from the right:

        >>> g[0]
        RectGrid(
            [-1.],
            [ 2.,  4.,  5.],
            [ 5.],
            [ 2.,  4.,  7.]
        )
        >>> g[0] == g[0, :, :, :] == g[0, ...]
        True
        """
        if isinstance(indices, list):
            if indices:
                new_coord_vecs = [self.coord_vectors[0][indices]]
                new_coord_vecs += self.coord_vectors[1:]
            else:
                new_coord_vecs = []
            return RectGrid(*new_coord_vecs)

        indices = normalized_index_expression(indices, self.shape,
                                              int_to_slice=False)

        # If all indices are integers, return an array (a point). Otherwise,
        # create a new grid.
        if all(np.isscalar(idx) for idx in indices):
            return np.fromiter(
                (v[int(idx)] for idx, v in zip(indices, self.coord_vectors)),
                dtype=float)
        else:
            new_coord_vecs = [vec[idx]
                              for idx, vec in zip(indices, self.coord_vectors)]
            return RectGrid(*new_coord_vecs)

    def __array__(self, dtype=None):
        """Used with ``numpy``. Returns `points`.

        This allows usage of RectGrid with some numpy functions.

        Parameters
        ----------
        dtype : `numpy.dtype`
            The Numpy data type of the result array. ``None`` means `float`.

        Examples
        --------
        >>> g = RectGrid([0, 1], [-2, 0, 2])

        Convert to an array:

        >>> np.asarray(g)
        array([[ 0., -2.],
               [ 0.,  0.],
               [ 0.,  2.],
               [ 1., -2.],
               [ 1.,  0.],
               [ 1.,  2.]])

        Calculate the midpoint:

        >>> np.mean(g, axis=0)
        array([ 0.5,  0. ])
        """
        return self.points().astype(dtype)

    def __repr__(self):
        """Return ``repr(self)``."""
        if self.is_uniform:
            ctor = 'uniform_grid'
            posargs = [self.min_pt, self.max_pt, self.shape]
            posmod = [array_str, array_str, '']
            with npy_printoptions(precision=4):
                inner_str = signature_string(posargs, [], mod=[posmod, ''])
            return '{}({})'.format(ctor, inner_str)
        else:
            ctor = self.__class__.__name__
            posargs = self.coord_vectors
            posmod = array_str
            inner_str = signature_string(posargs, [], sep=[',\n', ', ', ', '],
                                         mod=[posmod, ''])
            return '{}(\n{}\n)'.format(ctor, indent(inner_str))

    __str__ = __repr__


def uniform_grid_fromintv(intv_prod, shape, nodes_on_bdry=True):
    """Return a grid from sampling an interval product uniformly.

    The resulting grid will by default include ``intv_prod.min_pt`` and
    ``intv_prod.max_pt`` as grid points. If you want a subdivision into
    equally sized cells with grid points in the middle, use
    `uniform_partition` instead.

    Parameters
    ----------
    intv_prod : `IntervalProd`
        Set to be sampled.
    shape : int or sequence of ints
        Number of nodes per axis. Entries corresponding to degenerate axes
        must be equal to 1.
    nodes_on_bdry : bool or sequence, optional
        If a sequence is provided, it determines per axis whether to
        place the last grid point on the boundary (``True``) or shift it
        by half a cell size into the interior (``False``). In each axis,
        an entry may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``array.ndim``.

        A single boolean is interpreted as a global choice for all
        boundaries.

    Returns
    -------
    sampling : `RectGrid`
        Uniform sampling grid for the interval product.

    Examples
    --------
    >>> rbox = odl.IntervalProd([-1.5, 2], [-0.5, 3])
    >>> grid = uniform_grid_fromintv(rbox, (3, 3))
    >>> grid.coord_vectors
    (array([-1.5, -1. , -0.5]), array([ 2. ,  2.5,  3. ]))

    To have the nodes in the "middle", use ``nodes_on_bdry=False``:

    >>> grid = uniform_grid_fromintv(rbox, (2, 2), nodes_on_bdry=False)
    >>> grid.coord_vectors
    (array([-1.25, -0.75]), array([ 2.25,  2.75]))

    See Also
    --------
    uniform_grid : Create a uniform grid directly.
    odl.core.discr.partition.uniform_partition_fromintv :
        divide interval product into equally sized subsets
    """
    if not isinstance(intv_prod, IntervalProd):
        raise TypeError('{!r} is not an `IntervalProd` instance'
                        ''.format(intv_prod))

    if (np.any(np.isinf(intv_prod.min_pt)) or
            np.any(np.isinf(intv_prod.max_pt))):
        raise ValueError('`intv_prod` must be finite, got {!r}'
                         ''.format('intv_prod'))

    shape = normalized_scalar_param_list(shape, intv_prod.ndim, safe_int_conv)

    if isinstance(nodes_on_bdry, bool):
        nodes_on_bdry = ([(nodes_on_bdry, nodes_on_bdry)] *
                         intv_prod.ndim)
    elif intv_prod.ndim == 1 and len(nodes_on_bdry) == 2:
        nodes_on_bdry = [nodes_on_bdry]
    elif len(nodes_on_bdry) != intv_prod.ndim:
        raise ValueError('`nodes_on_bdry` has length {}, expected {}'
                         ''.format(len(nodes_on_bdry), intv_prod.ndim))
    else:
        shape = tuple(int(n) for n in shape)

    # We need to determine the placement of the grid minimum and maximum
    # points based on the choices in nodes_on_bdry. If in a given axis,
    # and for a given side (left or right), the entry is True, the node lies
    # on the boundary, so this coordinate can simply be taken as-is.
    #
    # Otherwise, the following conditions must be met:
    #
    # 1. The node should be half a stride s away from the boundary
    # 2. Adding or subtracting (n-1)*s should give the other extremal node.
    #
    # If both nodes are to be shifted half a stride inside,
    # the second condition yields
    # a + s/2 + (n-1)*s = b - s/2 => s = (b - a) / n,
    # hence the extremal grid points are
    # gmin = a + s/2 = a + (b - a) / (2 * n),
    # gmax = b - s/2 = b - (b - a) / (2 * n).
    #
    # In the case where one node, say the rightmost, lies on the boundary,
    # the condition 2. reads as
    # a + s/2 + (n-1)*s = b => s = (b - a) / (n - 1/2),
    # thus
    # gmin = a + (b - a) / (2 * n - 1).

    gmin, gmax = [], []
    for n, xmin, xmax, on_bdry in zip(shape, intv_prod.min_pt,
                                      intv_prod.max_pt, nodes_on_bdry):

        # Unpack the tuple if possible, else use bool globally for this axis
        try:
            bdry_l, bdry_r = on_bdry
        except TypeError:
            bdry_l = bdry_r = on_bdry

        if bdry_l and bdry_r:
            gmin.append(xmin)
            gmax.append(xmax)
        elif bdry_l and not bdry_r:
            gmin.append(xmin)
            gmax.append(xmax - (xmax - xmin) / (2 * n - 1))
        elif not bdry_l and bdry_r:
            gmin.append(xmin + (xmax - xmin) / (2 * n - 1))
            gmax.append(xmax)
        else:
            gmin.append(xmin + (xmax - xmin) / (2 * n))
            gmax.append(xmax - (xmax - xmin) / (2 * n))

    # Create the grid
    coord_vecs = [np.linspace(mi, ma, num, endpoint=True, dtype=np.float64)
                  for mi, ma, num in zip(gmin, gmax, shape)]
    return RectGrid(*coord_vecs)


def uniform_grid(min_pt, max_pt, shape, nodes_on_bdry=True):
    """Return a grid from sampling an implicit interval product uniformly.

    Parameters
    ----------
    min_pt : float or sequence of float
        Vectors of lower ends of the intervals in the product.
    max_pt : float or sequence of float
        Vectors of upper ends of the intervals in the product.
    shape : int or sequence of ints
        Number of nodes per axis. Entries corresponding to degenerate axes
        must be equal to 1.
    nodes_on_bdry : bool or sequence, optional
        If a sequence is provided, it determines per axis whether to
        place the last grid point on the boundary (``True``) or shift it
        by half a cell size into the interior (``False``). In each axis,
        an entry may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``array.ndim``.

        A single boolean is interpreted as a global choice for all
        boundaries.

    Returns
    -------
    uniform_grid : `RectGrid`
        The resulting uniform grid.

    See Also
    --------
    uniform_grid_fromintv :
        sample a given interval product
    odl.core.discr.partition.uniform_partition :
        divide implicitly defined interval product into equally
        sized subsets

    Examples
    --------
    By default, the min/max points are included in the grid:

    >>> grid = odl.uniform_grid([-1.5, 2], [-0.5, 3], (3, 3))
    >>> grid.coord_vectors
    (array([-1.5, -1. , -0.5]), array([ 2. ,  2.5,  3. ]))

    If ``shape`` is supposed to refer to small subvolumes, and the grid
    should be their centers, use the option ``nodes_on_bdry=False``:

    >>> grid = odl.uniform_grid([-1.5, 2], [-0.5, 3], (2, 2),
    ...                         nodes_on_bdry=False)
    >>> grid.coord_vectors
    (array([-1.25, -0.75]), array([ 2.25,  2.75]))

    In 1D, we don't need sequences:

    >>> grid = odl.uniform_grid(0, 1, 3)
    >>> grid.coord_vectors
    (array([ 0. ,  0.5,  1. ]),)
    """
    return uniform_grid_fromintv(IntervalProd(min_pt, max_pt), shape,
                                 nodes_on_bdry=nodes_on_bdry)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the smooth solvers."""

from __future__ import division
import odl


def test_backtracking_line_search():
    """Test some basic properties of BacktrackingLineSearch."""
    space = odl.rn(2)

    func = odl.functionals.L2NormSquared(space)

    line_search = odl.solvers.BacktrackingLineSearch(func)

    x = space.element([1, 0])
    for direction in [space.element([1, 0]),
                      space.element([-1, 0]),
                      space.element([-1, -1])]:
        dir_derivative = func.gradient(x).inner(direction)

        steplen = line_search(x, direction, dir_derivative)
        assert func(x + steplen * direction) < func(x)


def test_constant_line_search():
    """Test some basic properties of BacktrackingLineSearch."""
    space = odl.rn(2)

    func = odl.functionals.L2NormSquared(space)

    line_search = odl.solvers.ConstantLineSearch(0.57)

    x = space.element([1, 0])
    for direction in [space.element([1, 0]),
                      space.element([-1, 0]),
                      space.element([-1, -1])]:
        dir_derivative = func.gradient(x).inner(direction)

        steplen = line_search(x, direction, dir_derivative)
        assert steplen == 0.57


def test_line_search_from_iternum():
    """Test some basic properties of LineSearchFromIterNum."""
    space = odl.rn(2)

    func = odl.functionals.L2NormSquared(space)

    line_search = odl.solvers.LineSearchFromIterNum(lambda n: 1 / (n + 1))

    x = space.element([1, 0])
    for n, direction in enumerate([space.element([1, 0]),
                                   space.element([-1, 0]),
                                   space.element([-1, -1])]):
        dir_derivative = func.gradient(x).inner(direction)

        steplen = line_search(x, direction, dir_derivative)
        assert steplen == 1 / (n + 1)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""ODL (Operator Discretization Library).

ODL is a Python library for fast prototyping focusing on (but not
restricted to) inverse problems.
"""

from __future__ import absolute_import

from os import pardir, path, environ
environ['SCIPY_ARRAY_API']='1'

import numpy as np

__all__ = (
    'array_API_support'
    'set',
    'space',
    'operator',
    'discr',
    'contrib',
    'deform',
    'diagnostics',
    'phantom',
    'solvers',
    'tomo',
    'trafos',
    'ufunc_ops',
)

# Set package version
curdir = path.abspath(path.dirname(__file__))

with open(path.join(curdir, 'VERSION')) as version_file:
    __version__ = version_file.read().strip()

# Set old Numpy printing behavior as to not invalidate all doctests.
# TODO(kohr-h): switch to new behavior when Numpy 1.14 is minimum
try:
    np.set_printoptions(legacy='1.13')
except TypeError:
    pass

# Set printing line width to 71 to allow method docstrings to not extend
# beyond 79 characters (2 times indent of 4)
np.set_printoptions(linewidth=71)

# Import all names from "core" subpackages into the top-level namespace;
# the `__all__` collection is extended later to make import errors more
# visible (otherwise one gets errors like "... has no attribute __all__")
from .core.set import __all__ as _core_set_all
from .core.set import *
from .core.array_API_support import __all__ as _core_arrayAPI_all
from .core.array_API_support import *
from .core.discr import __all__ as _core_discr_all
from .core.discr import *
from .core.operator import __all__ as _core_operator_all
from .core.operator import *
from .core.space import __all__ as _core_space_all
from .core.space import *

# More "advanced" subpackages keep their namespaces separate from top-level,
# we only import the modules themselves
from . import contrib

from .core import diagnostics
diagnostics.__module__ = "odl"

from .core import phantom
phantom.__module__ = "odl"

from . import solvers
from . import functionals
# from .applications import tomo
from . import trafos
# from . import ufunc_ops

# Add `test` function to global namespace so users can run `odl.test()`
from .core.util import test

# Make often-used ODL definitions appear as members of the main `odl` namespace
# in the documentation (they are aliased in that namespace), even though they
# are defined in modules with more verbose names.
for module_source in [_core_set_all, _core_arrayAPI_all, _core_discr_all, _core_operator_all, _core_space_all]:
    for entity in module_source:
        globals()[entity].__module__ = "odl"
    del entity
del _core_set_all, _core_arrayAPI_all, _core_discr_all, _core_operator_all, _core_space_all



# Amend `__all__`
__all__ += ('test',)



# coding=utf-8

# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for the d.c. solvers."""

from __future__ import division
import odl
from odl.solvers import dca, prox_dca, doubleprox_dc
from odl.solvers.nonsmooth.difference_convex import doubleprox_dc_simple
import numpy as np
import pytest


# Places for the accepted error when comparing results
HIGH_ACCURACY = 8
LOW_ACCURACY = 4


def test_dca():
    """Test the dca method for the following simple problem:

    Let a > 0, and let b be a real number. Solve

    min_x a/2 (x - b)^2 - |x|

    For finding possible (local) solutions, we consider several cases:
    * x > 0 ==> ∂|.|(x) = 1, i.e., a necessary optimality condition is
      0 = a (x - b) - 1 ==> x = b + 1/a. This only happens if b > -1/a.
    * x < 0 ==> ∂|.|(x) = -1, i.e., a necessary optimality condition is
      0 = a (x - b) + 1 ==> x = b - 1/a. This only happens if b < 1/a.
    * x = 0 ==> ∂|.|(x) = [-1, 1], i.e., a necessary optimality condition is
      0 = a (x - b) + [-1, 1] ==> b - 1/a <= x = 0 <= b + 1/a.

    To summarize, we might find the following solution sets:
    * {b - 1/a} if b < -1/a,
    * {-2/a, 0} if b = -1/a,
    * {b - 1/a, 0, b + 1/a} if -1/a < b < 1/a,
    * {0, 2/a} if b = 1/a,
    * {b + 1/a} if b > 1/a.
    """

    # Set the problem parameters
    a = 0.5
    b = 0.5
    # This means -1/a = -2 < b = 0.5 < 1/a = 2.
    space = odl.rn(1)
    f = a / 2 * odl.functionals.L2NormSquared(space).translated(b)
    g = odl.functionals.L1Norm(space)
    niter = 50

    # Set up some space elements for the solvers to use
    x = space.element(-0.5)
    x_dca = x.copy()
    x_prox_dca = x.copy()
    x_doubleprox = x.copy()
    x_simpl = x.copy()

    # Some additional parameters for some of the solvers
    phi = odl.functionals.ZeroFunctional(space)
    y = space.element(3)
    y_simpl = y.copy()
    gamma = 1
    mu = 1
    K = odl.IdentityOperator(space)

    dca(x_dca, f, g, niter)
    prox_dca(x_prox_dca, f, g, niter, gamma)
    doubleprox_dc(x_doubleprox, y, f, phi, g, K, niter, gamma, mu)
    doubleprox_dc_simple(x_simpl, y_simpl, f, phi, g, K, niter, gamma, mu)
    expected = np.asarray([b - 1 / a, 0, b + 1 / a])

    dist_dca = np.min(np.abs(expected - float(x_dca)))
    dist_prox_dca = np.min(np.abs(expected - float(x_prox_dca)))
    dist_prox_doubleprox = np.min(np.abs(expected - float(x_doubleprox)))

    # Optimized and simplified versions of doubleprox_dc should give
    # the same result.
    assert float(x_simpl) == pytest.approx(float(x_doubleprox))
    assert float(y_simpl) == pytest.approx(float(y))

    # All methods should give approximately one solution of the problem.
    # For 50 iterations, the methods have been tested to achieve an absolute
    # accuracy of at least 1/10^6.
    assert float(dist_dca) == pytest.approx(0, abs=1e-6)
    assert float(dist_prox_dca) == pytest.approx(0, abs=1e-6)
    assert float(dist_prox_doubleprox) == pytest.approx(0, abs=1e-6)
